/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
/* evolutionary parameters */
	#define EVOL_C
	#include "headers.h"

	#define ALPHA 2.0            /* exponential fitness scale, unused now */
	#define WATCH_COUNT 10000

	#define SQRT2 1.41421356237310

	/* linear approximation to erf(x), for speed */
	REAL myerf(REAL x) {
	  int i;
	  REAL ret;
	  REAL frac;
	  REAL whole;
	  static float mem[101];
	  static int first=1;
	  REAL sign=1.0;
	  if (first) {
	    for (i=0;i<=100;i++)
	      mem[i]=erf( (4.0 / 100.0)*i);
	    first = 0;
	  }
	  if (x == 0.0) return 0.0;
	  if (x < 0.0) {x = -x; sign = -1.0; }
	  if (x >= 4.0) ret = 1.0;
	  else {
	    whole = x * (100.0 / 4.0);
	    i = whole;
	    frac = whole - i;
	    /*      printf("i %d x %7.6f frac %7.6f memi %7.6f memi+1 %7.6f\n",
		    i,x,frac,mem[i],mem[i+1]); */
	    ret = mem[i]*(1.0 - frac)+(mem[i+1])*frac;
	  }
	  return sign * ret;
	}

	#ifdef PARALLEL

	int hasforked=0;
	int haslocked OFE;

	void my_lock() {
	  USEMINE
	    if (hasforked && (haslocked MINE == 0)) m_lock();
	  haslocked MINE ++;
	#ifdef CHECKS
	  if (haslocked MINE > 5) printf("%d:too many locks\n",m_get_myid());
	#endif
	}

	void my_unlock() {
	  USEMINE
	    haslocked MINE --;
	  if (hasforked && (haslocked MINE ==0)) m_unlock();
	#ifdef CHECKS
	  if (haslocked MINE < 0) printf("%d:lock underflow\n",m_get_myid());
	#endif
	}

	#endif


	typedef struct {
	  xybrickptr *roots;
	  int rootcount;
	  brick *bricks;
	  int brickcount;
	  REAL fitness;
	  REAL mass;
	#ifdef PARALLEL
	  unsigned long lock;
	#endif
	} individual;

	individual population[POP_SIZE];

	REAL maxfitness = -1.0E32,
	  minfitness = 1.0E32,
	  sumfitness=0.0,
	  sumsqfitness=0.0;

	#ifdef PARALLEL
	lock_ind(int i, int lock) {
	  test_then_add(&(population[i].lock),lock);
	}
	#endif

	void update_population(void);

	void adopt_all() {
	  int i,k,j=0,dad=0,r;

	  for (i=0;i<brickcount;i++) {
	    if (parent[i] < 0) {
	      /* it's an orphan, try to find an adoptive parent */
	      for (k=0;k<adoptcount;k++)
		if (adoptives[k].son == i &&
		    (dad=adoptives[k].father,
		     j=free_joint(&bricks[dad])) >=0)
		  break;
	      if (k<adoptcount) {
		/* adopt him */
		parent[i] = dad;
		bricks[dad].joints[j]=i;
		bricks[dad].jointpos[j].x = brickpos[i].x - brickpos[dad].x;
		bricks[dad].jointpos[j].y = brickpos[i].y - brickpos[dad].y;
		/* delete k-th entry from adoptives list */
		while (k+1 < adoptcount) {
		  adoptives[k]=adoptives[k+1];
		  k++;
		}
		adoptcount--;
		/* delete this moron from roots list: ain't no root anymore */
		for (r=0;r<rootcount;r++)
		  if (roots[r].no==i) break;
		if (roots[r].no==i) {
		  while ((r+1) < rootcount) {
		    roots[r]=roots[r+1];
		    /* OBJCPY(roots[r],roots[r+1]); */
		    r++;
		  }
		  rootcount--;
		}

	      }
	    }
	  }
	}

	void start_pop() {
	  USEMINE

	  popcount = 0;

	  /* WARNING: In this version we assume that the model
	     is already loaded into the buffer and it's not too heavy at all.
	     Here we just (carelessly) evaluate fitness and procceed */

	  fitnessvalue MINE = fitness();
	  update_population();
	}


	#define MMS(x,ma,mi,su) su += x; if(x > ma) ma = x; if(x < mi) mi = x;

	void watch(count) {
	  USEMINE
	    int i,best=-1,worst=-1;
	  int minbricks,maxbricks,sumbricks;
	  REAL minfit,maxfit,sumfit;
	  REAL minmass,maxmass,summass;
	  static int elapsed = 0;
	  static int lastringcount = 1;
	  static FILE *fp = NULL;

	  WHOAM2
	    fprintf(stderr,"watch %d\n",count);
	  fflush(stderr);
	  if (fp==NULL) fp=fopen("watch","w");
	  minbricks = 100;
	  maxbricks = -100;
	  sumbricks = 0;
	  minfit = minmass = 1e32;
	  maxfit = maxmass = -1e32;
	  sumfit = summass = 0.0;
	  fprintf (fp,"%6d %d ",count,popcount);
	  for (i=0;i<popcount;i++) {
	#ifdef PARALLEL
	    if ((population[i].lock & WRITE_LOCK)==0) {
	#endif
	      MMS(population[i].brickcount,maxbricks,minbricks,sumbricks);
	      MMS(population[i].fitness,maxfit,minfit,sumfit);
	      MMS(population[i].mass,maxmass,minmass,summass);
	      if (population[i].fitness == maxfit) best=i;
	      if (population[i].fitness == minfit) worst=i;
	#ifdef PARALLEL
	    }
	#endif
	  }
	  fprintf (fp,"fi %f %f %f br %d %d %f ma %f %f %f ",
		   maxfit,minfit,sumfit / popcount,
		   maxbricks,minbricks, (REAL) sumbricks / popcount,
		   maxmass,minmass,summass / popcount);
	  initgrid();
	  brickcount MINE=0;

	  if(loadmodel(population[best].roots,population[best].rootcount,
		       population[best].bricks)
	     || makemodel() || tooheavy()) {

	    dumpfile("err",
		     population[best].roots,population[best].rootcount,
		     population[best].bricks, population[best].brickcount,
		     bodies MINE, bodycount MINE,
		     joints MINE, jointcount MINE);
	    abortonerror(3,"can't load best");
	  }
	  mbricks("BEST.m",
		  roots, rootcount,
		  bricks MINE,brickcount MINE,fitness(),
		  bodies MINE,bodycount MINE,joints MINE,jointcount MINE);
	  initgrid();
	  brickcount MINE=0;
	  brickcount MINE=0;
	  loadmodel(population[worst].roots,population[worst].rootcount,
		    population[worst].bricks);
	  makemodel();
	  tooheavy();
	  mbricks("WORST.m",
		  roots,rootcount,
		  bricks MINE,brickcount MINE,fitness(),
		  bodies MINE,bodycount MINE,joints MINE,jointcount MINE);

	  i=time(NULL);
	  if (elapsed) fprintf(fp,"sec %d rc %ld ",i-elapsed,ringcount-lastringcount);
	  if (elapsed) fprintf(stderr,"=%d secs.",i-elapsed);
	  lastringcount=ringcount;
	  elapsed=i;
	  fprintf(fp,"\n");
	  fflush(fp);

	}



	void save_ind_at(int i) {
	  USEMINE

	#if (defined PARALLEL) && (defined CHECKS)
	    if (population[i].lock != WRITE_LOCK)
	      printf("warning: save at %d lock %x\n",i,population[i].lock);
	#endif
	  ARRALLOC(population[i].bricks,brickcount);
	  ARRCPY(population[i].bricks,bricks,brickcount);
	  ARRALLOC(population[i].roots,rootcount);
	  ARRCPY(population[i].roots,roots,rootcount);
	  population[i].brickcount = brickcount MINE;
	  population[i].rootcount = rootcount;
	  population[i].fitness = fitnessvalue MINE;
	  population[i].mass = totalmass MINE;

	#ifdef PRINTS
	  printf("SA %d bc=%d f=%f m=%f\n",
		 i,brickcount MINE,fitnessvalue MINE, totalmass MINE);

	  for (i=0;i<brickcount MINE;i++) {
	    printf("brick %d size %d joints ",
		   i,
		   bricks MINE[i].size);
	    {
	      int j;
	      for (j=0; j<4; j++) {

		printf("%d(%d,%d) ",bricks MINE [i].joints[j],
		       bricks MINE [i].jointpos[j].x,bricks MINE [i].jointpos[j].y);
	      }
	    }
	    printf("\n");
	  }
	  printf("-----------\n");

	#endif
	}

	void recalc_min_max() {
	  int i;
	  REAL x;

	  minfitness=1E32;
	  maxfitness=1E-32;
	  for (i=0;i<popcount;i++) {
	    x = population[i].fitness;
	    if (x < minfitness) minfitness=x;
	    if (x > maxfitness) maxfitness=x;
	  }
	}

	int select_on_fitness(int inverse /*REAL alpha*/ LOCK_PARAM) {
	  REAL x,acc;
	  REAL stdev,avg,phi;
	  int i;
	  static int start=0;

	  /* phi(m,s,x) = P(X<=x) where X is N(m,s^2) */

	#define PHI(m,s,x) ((1+myerf(((x)-(m))/(SQRT2*(s))))/2)

	#ifdef PARALLEL
	  int done;
	#endif

	  if (start > popcount) start=0;

	#ifdef PARALLEL
	  done=0;
	  while (!done) {
	#endif
	    if (minfitness == maxfitness)
	      i=RANDINT() % popcount;
	    else {
	      avg = sumfitness / popcount ;
	      stdev = sumsqfitness/popcount - avg*avg;
	      x = ((REAL) (RANDINT() % 1000000) / 1000000.0 )* 0.5;
	      /* x is a random number in [0,popcount/2)
		 0.5 is E(Phi(X)), ergo popcount/2 is E(SUM(Phi(X))) */
	      x = x * popcount;
	      acc = 0.0;
	      for (i=start; acc < x ;i=((i+1)%popcount)) {
		phi = PHI(avg,stdev,population[i].fitness);
		if (inverse) phi = 1-phi;
		acc += phi;
		if (acc > x) break;
	      }
	    }
	#ifdef PARALLEL
	  }
	#endif

	#ifdef CHECKS
	  if (i<0 || i>= popcount) {
	    FILE *fp;
	    printf("out of bounds in select_on_fitness\n");
	    fp=fopen("dump","w");
	    printf("dumping fitnesses to dump\n");
	    for (i=0;i<popcount;i++)
	      fprintf(fp,"%5d %14.7f\n",i,population[i].fitness);
	    fclose(fp);
	  }
	#endif
	  return i;
	}



	int select_random(/*REAL alpha*/ LOCK_PARAM) {
	  int i;

	#ifdef PARALLEL
	  int done;
	#endif

	#ifdef PARALLEL
	  done=0;
	  while (!done) {
	#endif
	    i=(RANDINT() % popcount);
	#ifdef PARALLEL
	    if (i<popcount) {
	      LOCK
		if (to_write) {
		  if (population[i].lock == 0) {
		    population[i].lock = WRITE_LOCK;
		    done=1;
		  }
		}
		else {
		  if ((population[i].lock & WRITE_LOCK)==0) {
		    population[i].lock++;
		    done=1;
		  }
		}
	      UNLOCK
		}
	  }
	#else
	#ifdef CHECKS
	  if (i<0 || i>= popcount) {
	    FILE *fp;
	    printf("out of bounds in select_on_fitness\n");
	    fp=fopen("dump","w");
	    printf("dumping fitnesses to dump\n");
	    for (i=0;i<popcount;i++)
	      fprintf(fp,"%5d %14.7f\n",i,population[i].fitness);
	    fclose(fp);
	  }
	#endif
	#endif
	  return i;
	}

	void update_population(void) {
	  USEMINE
	    int i;
	  int recalc;
	  int newmax;
	#ifndef CGI
	  int j;
	  char s[50],t[50];
	#endif
	  static unsigned long count=0;
	  unsigned long mycount;
	  REAL x;
	#ifdef PARALLEL
	  int locked;
	#endif


	  mycount = INCREMENT(count);
	  recalc=newmax=0;
	#ifdef MAKELINEAGE
	  if (lineage==NULL) lineage=fopen("lineage","w");
	#endif
	#ifdef SAVELINEAGE
	  if (lineage==NULL) {
	    lineage=fopen("linlist","r");
	    fscanf(lineage,"%d %d %d",&nextsave,&i,&j);
	    printf("next save at %d\n",nextsave);
	  }
	#endif

	#ifdef PARALLEL
	  if (popcount < POP_SIZE) {
	    LOCK
	      locked=1;
	  }
	  else locked=0;
	#endif


	  if (popcount == POP_SIZE) {
	#ifdef PARALLEL
	    if (locked) UNLOCK
	#endif

			  i=select_on_fitness(/*-ALPHA*/1 TO_WRITE);
	    x=population[i].fitness;

	    if (x <= minfitness || x >= maxfitness) recalc=1;
	    sumfitness -= x;
	    sumsqfitness -= (x*x);
	    free(population[i].bricks);
	    free(population[i].roots);
	  }
	  else {
	    i=popcount++;
	#ifdef PARALLEL
	    population[i].lock = WRITE_LOCK;
	    population[i].fitness = minfitness;
	    UNLOCK
	#endif
	      }

	  if (fitnessvalue < minfitness) minfitness=fitnessvalue;
	  if (fitnessvalue > maxfitness) {
	    maxfitness=fitnessvalue;
	    newmax=1;
	  }
	  sumfitness += fitnessvalue;
	  sumsqfitness += (fitnessvalue * fitnessvalue);

	#ifdef MAKELINEAGE
	  fprintf(lineage,"%d %d %d\n",count,parent,i);
	  fflush(lineage);
	#endif
	  save_ind_at(i);
	#ifdef PARALLEL
	  test_then_add(&(population[i].lock),-WRITE_LOCK);
	#endif
	#ifdef SAVELINEAGE
	  if (mycount >= nextsave) {
	    printf("saving %d %d\n",mycount,nextsave);
	    sprintf(t,"%5d",savecount++);
	    for (i=0;t[i]!='\0';i++) if (t[i]==' ' ) t[i]='0';
	    sprintf(s,"LIN%s.m",t);
	    mbricks(s,bricks MINE,brickcount MINE,fitnessvalue MINE,
		    bodies MINE, bodycount MINE, joints MINE, jointcount MINE);

	    fscanf(lineage,"%d %d %d",&nextsave,&j,&k);
	    printf("next save at %d\n",nextsave);

	  }
	#endif

	  count++;
	  if (recalc) recalc_min_max();

	#ifndef CGI
	  if (newmax) {
	    LOCK
	      if (fitnessvalue MINE >= maxfitness) {
		maxfitness=fitnessvalue MINE;
		sprintf(t,"%7d",(int) (fitnessvalue MINE*10000));
		for (j=0;t[j]!=0;j++) if (t[j]==' ') t[j]='0';
		WHOAM2
		  fprintf(stderr,"MAX %s\n",t);
		sprintf(s,"MAX%s.m",t);

		mat_bricks(s);
		unlink("MAX.m");
		link(s,"MAX.m");
		sprintf(s,"MAX%s.info",t);
		dump_bricks(s);
		unlink("MAX.info");
		link(s,"MAX.info");
	#ifdef VBLOCKS
		sprintf(s,"MAX%s.vb",t);
		vb_bricks(s,bricks,brickcount,fitnessvalue);
		unlink("MAX.vb");
		link(s,"MAX.vb");
	#endif
		UNLOCK
		  }
	  }
	#endif
	}

	unsigned long compt =0, compf = 0, mutat = 0, mutaf = 0, crost = 0, crosf=0;

	int parent_tree OFE;

	void evolve() {
	  USEMINE
	    int i,j;
	#ifdef PARALLEL
	  static unsigned long bigcount=0;
	#endif
	  int status;
	  int count;
	  int timelimit=30;
	  time_t timeout;

	#ifdef PARALLEL
	  count=INCREMENT(bigcount);
	#ifdef PRINTS
	  if(myid==0) rseed=SRANDOM;
	  m_sync();
	  printf("(%d) init evolve count=%d\n",myid,count);
	  i=myid+rseed;
	  RANDSEED(i);
	  printf("(%d) init randseed=%d\n",myid,i);
	#endif
	#else
	  count=0;
	#endif

	  timeout = time(NULL) + timelimit;

	  while (maxfitness < TARGET_FITNESS && count < COUNTLIMIT) {
	    if ((count % 1000) ==0 && time(NULL) > timeout) break;
	#if defined(PRINTS) && defined(PARALLEL)
	    printf("(%d)evolve loop max=%f count=%d\n",myid,maxfitness,count);
	#endif
	    if ((!NOCROSSOVER) && RANDINT() % 2) {
	      /* crossover */
	      i=select_on_fitness(0 /*ALPHA*/ TO_READ);
	      j=select_random(0 /*ALPHA*/ TO_READ);
	#ifdef PRINTS
	      printf("co %d %d ",i,j);
	#endif
	      parent_tree MINE=i;
	      INCREMENT(crost);
	      status=crossover(
			       population[i].roots,population[i].rootcount,
			       population[i].bricks,population[i].brickcount-PREDEF_BRICK_NO,
			       population[j].roots,population[j].rootcount,
			       population[j].bricks,population[j].brickcount-PREDEF_BRICK_NO);

	#ifdef PARALLEL
	      test_then_add(&(population[i].lock),-1);
	      test_then_add(&(population[j].lock),-1);
	#endif
	      if (status==0 &&
		  ((popcount < POP_SIZE) || (fitnessvalue >= minfitness)))
		update_population();
	      else {
		INCREMENT(crosf);
	#ifdef PRINTS
		printf("co fail\n");
	#endif
	      }
	    }
	    else {
	      /* mutate */
	      {
		/* plain mutation */
		i=select_on_fitness(0 /*ALPHA*/ TO_READ);
	#ifdef PRINTS
		printf("mut %d ",i);
		if (i==1022)
		  printf("here we are\n");
	#endif
		parent_tree MINE=i;
		INCREMENT(mutat);
		status = mutate(population[i].roots,population[i].rootcount,
				population[i].bricks,
				population[i].brickcount-PREDEF_BRICK_NO);

	#ifdef PARALLEL
		test_then_add(&(population[i].lock),-1);
	#endif
		if (status==0 &&
		    ((popcount < POP_SIZE) || (fitnessvalue >= minfitness)))
		  update_population();
		else {
		  INCREMENT(mutaf);
	#ifdef PRINTS
		  printf("mut fail\n");
	#endif
		}
	      }
	    }
	#ifdef PARALLEL
	    count=INCREMENT(bigcount);

	    if ((count % WATCH_COUNT) == 0) {
	      LOCK
		watch(count);
	      fflush(stdout);
	      UNLOCK
		};
	    /*

	      if ((i=m_next()) >= WATCH_COUNT-PROCESSORS) {
	      #ifdef CHECK
	      if (haslocked[myid]) {
	      printf("panic: goint into sync with lock\n");
	      }
	      #endif
	      m_sync();
	      if (myid == 0) watch(count);
	      m_sync();
	      }
	    */
	#else
	    count++;
	    if ((count % WATCH_COUNT)==0) {
	      watch(count);
	      fflush(stdout);
	    }
	#endif
	  }
	}

	void save_best(char *name) {
	  double maxfit;
	  int i;
	  int best;
	  maxfit=-1e32;
	  best=0;
	  for (i=0;i<popcount;i++)
	    if (population[i].fitness > maxfit) {
	      best=i;
	      maxfit=population[i].fitness;
	    }
	  initgrid();
	  brickcount=0;
	  if(loadmodel(population[best].roots,population[best].rootcount,
		       population[best].bricks)
	     || makemodel() || tooheavy()) {

	    dumpfile("err.info",
		     population[best].roots,population[best].rootcount,
		     population[best].bricks, population[best].brickcount,
		     bodies MINE, bodycount MINE,
		     joints MINE, jointcount MINE);
	    abortonerror(7,"cant load best");
	  }
	  mat_bricks(name);
	  save_problem(stdout);
	  fprintf(stderr,"\n-------\nevol.cgi returns: \n");
	  save_problem(stderr);
	  fflush(stderr);
	}

	int main(int argc,char **argv) {
	  int i;

	#ifdef PARALLEL
	  int myid=0;
	#endif

	#ifdef CGI
	  if (getenv("REQUEST_METHOD") != NULL) {
	    /* I'm running as CGI, output header */
	    printf("content-type: text/plain\n\n");
	  }
	  freopen("debug_evol","w",stderr);
	#endif

	#ifdef USEALARM
	  itrings();
	  ringing=0;
	#endif

	#ifdef PARALLEL
	  for (i=0;i<PROCESSORS;i++) {
	    haslocked[i]=0;
	  }
	#endif

	  load_problem(stdin);

	#ifdef PRINTS
	  save_file_after_parse = 1;
	#endif
	  convert_problem();
	  simulation_mode = EMBEDDED_MODE;

	  start_pop();
	  minfitness=1e32;
	  maxfitness=-1e32;

	  i=SRANDOM;
	  fprintf(stderr,"fit (%s) starts with seed %d\n",argv[0],i);
	  RANDSEED(i);
	#ifdef PARALLEL
	  m_set_procs(PROCESSORS);
	  hasforked=1;
	  m_fork(evolve);
	#else
	  evolve();
	#endif
	  /* save best */
	  save_best("last.m");
	  exit(0);
	}

	void xymain() {
	  int x,y,z;
	  initgrid();
	  while (1) {
	    scanf("%d %d",&x,&y);
	    printf("%d %d = %d\n",x,y,getgrid(x,y));
	    printf("new value-> ");
	    scanf("%d",&z);
	    printf("%d %d (%d) = %d == ",x,y,setgrid(x,y,z),z);
	    printf("%d \n",getgrid(x,y));
	  }
	}


