/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#define EXPLICIT_C

#include "headers.h"

  struct stack_struct {
    body *b;
    REAL frac;
  } ;


REAL flow(body *bodies, int bodycount,
	  joint *joints, int jointcount,
	  struct appliedforce *forces,
	  int forcecount) {
  int i,j;
  body *body;
  joint *joint;
  int stacktop=0;
  REAL jointsum;
  REAL frac,curfrac;
  REAL sign;
  REAL tau;
  REAL ret;
  static struct stack_struct stack[MAXBODIES+1];

  /* normalize weights */

  for (i=0;i<bodycount;i++) {
    body=&bodies[i];
    jointsum=0;
    for (j=0;j<body->jointcount;j++) {
#ifdef CHECKS
      if (body->joints[j]->weight <=0.0 || isnan(body->joints[j]->weight)) {
	fprintf(stderr,"warning: zero value on joint %d=%f\n",j,
		body->joints[j]->weight);
      }
#endif
      jointsum += body->joints[j]->weight *
	body->joints[j]->maxtorque;
      }
    if (body->jointcount == 0)
      return 0.0; /* a body is in the air, no flow */
    if (jointsum==0.0) {
      fprintf(stderr,"Warning: jointsum==0\n");
      for (j=0;j<body->jointcount;j++)
	body->joints[j]->weight =
	  1.0 / body->jointcount ;
    }
    else {
      for (j=0;j<body->jointcount;j++)
	body->joints[j]->weight =
	  body->joints[j]->weight/jointsum;
    }
  }

  for (j=0;j<jointcount;j++) {
    joints[j].torques = 0.0;
    joints[j].forces.x =
      joints[j].forces.y = 0.0;
  }

  for (i=0;i<forcecount MINE;i++) {
    stack[stacktop].b = &bodies[forces[i].target];
    stack[stacktop].frac = 1.0;
    stacktop = 1;
    while (stacktop > 0) {
      stacktop--;
      body = stack[stacktop].b;
      frac = stack[stacktop].frac;
      for (j=0; j<body->jointcount;j++) {
	joint=body->joints[j];
	sign=body->jointsign[j];
	tau = torqueat(joint->x,forces[i].x,forces[i].f);
	curfrac = frac * joint->weight * joint->maxtorque;
	joint->torques += tau * sign * curfrac ;
	if (NULL != body->joins[j]) {
	  stack[stacktop].b = body->joins[j];
	  stack[stacktop].frac = curfrac;
	  stacktop++;
	}
      }
    }
  }
  /* all torques have been computed, get fractions */
  ret = 1E32;
  avgload=0.0;
  for (i=0;i<jointcount;i++) {
    jointsum = fabs(joints[i].maxtorque / joints[i].torques) ;
    if (jointsum > 1.0)
      avgload += 1.0;
    else
      avgload += jointsum;
    if (ret > jointsum) ret=jointsum;
  }
  avgload = avgload / jointcount;
  /* now ret contains the minimum supported fraction */
  return ret;
}


