/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#define INTERFACE_C

#include "headers.h"
#include <ctype.h>

/**
 * Retrieve next nonempty, non-#-commented,
 * line, removing \n character at the end if needed.
 */

char *fgets_noeox(char *s, int size, FILE *fp) {
	char *p;
	int l;
	/* skip empty and/or # lines */
	do {
		p=fgets(s, size, fp);
		if (NULL==p)
			abortonerror(111, "Unexpected EOF");
	} while (s[0] == '\0' || s[0] == '\n' || s[0]=='#');
	l=strlen(s);
	if (l > 0 && s[l-1]=='\n')
		s[l-1]='\0';
	return s;
}

int *tokenize(char *s) {
#define MAX_TOKENS 100
	static int tokens[MAX_TOKENS+1];
	int state=0;
	tokens[0] = 0;
	for (; tokens[0]<MAX_TOKENS &&*s != 0; s++)
		switch (state) {
		case 1:
			if (isdigit(*s))
				tokens[tokens[0]]=(tokens[tokens[0]]*10 + (*s) - '0');
			else
				state=0;
			break;
		case 0:
			if (isdigit(*s)) {
				state=1;
				tokens[0]++;
				tokens[tokens[0]] = (*s) - '0';
			}
			break;
		}
	return tokens;
}

void load_problem(FILE *fp) {
	int i, kind, x, y, size;
	int *tokens;
	char s[1000];
	FILE *deb;

	deb=fopen("tmp/debug_load", "w");

	do {
		fgets_noeox(s, sizeof s, fp);
		if (deb) {
			fprintf(deb, "%s\n", s);
			fflush(deb);
		}
		tokens=tokenize(s);
		if (tokens[0]!=1)
			switch (tokens[1]) {
			case 100:
				minx=tokens[2];
				maxx=tokens[3];
				miny=tokens[4];
				maxy=tokens[5];
				break;
			default:
				abortonerror(106, "wrong entry <%s>", s);
			}
	} while (tokens[0] != 1);

	obj_count = tokens[1];
	if (obj_count <= 0) {
		abortonerror(101, "obj count < 0");
	}

	for (i=0; i<obj_count; i++) {
		fgets_noeox(s, sizeof s, fp);
		if (deb) {
			fprintf(deb, "%s\n", s);
			fflush(deb);
		}
		tokens=tokenize(s);
		if (4 != tokens[0]) {
			abortonerror(102, "wrong input format");
		}
		kind=tokens[1];
		x=tokens[2];
		y=tokens[3];
		size=tokens[4];
		if (kind < 0 || kind > KINDS_COUNT)
			abortonerror(104, "wrong obj kind");
		if (x < minx || (kind != LOAD_KIND && (x + size > maxx)) || y < miny
				|| y >= maxy) {
			abortonerror(105, "wrong obj position ");
		}
		objects[i].kind=kind;
		objects[i].x=x;
		objects[i].y=y;
		objects[i].size = size;
		objects[i].id=i;
	}

	/* sort objects by type */
	/* q&d bubble */
	{
		obj_type obj;
		int j;
		for (i=1; i<obj_count; i++) {
			/* find new object location for object i */
			j=i;
			while (j>0 && objects[j].kind < objects[j-1].kind) {
				obj=objects[j];
				objects[j]=objects[j-1];
				objects[j-1]=obj;
				j--;
			}
		}
	}
	/* locate kind boundaries */
	for (kind=0; kind<KINDS_COUNT; kind++) {
		object_kind_count[kind]=0;
		object_kind[kind]=NULL;
	}
	kind=0;
	x=0;
	object_kind[objects[0].kind]=&objects[0];
	for (i=0; i<obj_count; i++) {
		if (objects[i].kind != kind) {
			object_kind_count[kind]=i-x;
			kind=objects[i].kind;
			x=i;
			object_kind[kind]=&objects[i];
		}
	}
}

void save_problem(FILE *fp) {
	int i;
	int count;

#ifdef CGI_OLDSTYLE
#define SEP ' '
#else
#define SEP '\n'
#endif
	/* count non-evolvable objects */
	for (i=0, count=0; i<obj_count; i++)
		if (objects[i].kind != BRICK_KIND)
			count++;
	/* total objects = count + brickcount */
	fprintf(fp, "%d%c", count+brickcount, SEP);
	/* dump non-evolvable objects */
	for (i=0; i<obj_count; i++)
		if (objects[i].kind != BRICK_KIND) {
			fprintf(fp, "%d %d %d %d%c", objects[i].kind, objects[i].x,
					objects[i].y, objects[i].size, SEP);
		}
	/* dump bricks */
	for (i=0; i<brickcount; i++)
		fprintf(fp, "%d %d %d %d%c", BRICK_KIND, brickpos[i].x, brickpos[i].y,
				bricks[i].size, SEP);
	/* done ! */
}

int load_objects_conditional() {
	int k;
	int ret;
	xybrickptr *myroots;
	brick *mybricks;
	int myrootcount;

	for (k=0, myrootcount=0; k<obj_count; k++)
		if (objects[k].kind == BRICK_KIND && still_alive[k])
			myrootcount++;

	ARRALLOC(myroots,myrootcount);
	ARRALLOC(mybricks,myrootcount);

	for (k=0, myrootcount=0; k<obj_count; k++) {
		/* initialize everybody as myroots, at first */
		if (objects[k].kind == BRICK_KIND && still_alive[k]) {
			mybricks[myrootcount]=newbrick(objects[k].size);
			myroots[myrootcount].x=objects[k].x;
			myroots[myrootcount].y=objects[k].y;
			myroots[myrootcount].no=myrootcount;
			myrootcount++;
		}
	}

	if (loadmodel(myroots, myrootcount, mybricks))
		ret = -1;

	else if (makemodel() )
		ret = -2;

	else
		ret=0;
	free(myroots);
	free(mybricks);
	return ret;
}

int make_object_model_conditional() {
	int ret;
	int i;
#ifdef PRINTS
	printf("MO ");
#endif
	initgrid();
	jointcount = 0;
	bodycount = 0;
	brickcount = 0;
	rootcount = 0;
	totalmass = 0.0;
	adoptcount=0;
	ret = 0;
	for (i=0; i<obj_count; i++) {
		if (still_alive[i] && objects[i].kind == BRICK_KIND) {
			bricks[brickcount] = newbrick(objects[i].size);
			if (addbrick(objects[i].x, objects[i].y, brickcount, GROUND)) {
				ret = 1;
				break;
			} else {
				roots[rootcount].x = objects[i].x;
				roots[rootcount].y = objects[i].y;
				roots[rootcount].no = brickcount;
				brickcount++;
				rootcount++;
			}
#ifdef PRINTS
			printf("ab000=%d ", ret);
#endif
		}
	}

#ifdef PRINTS
	if (ret)
		printf("MO Failed\n");
	else
		printf("\n");
#endif
	return ret;
}

