/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#include "headers.h"

/* #define PRINTS */


void pprn2_(int *ln, char *name, int *exit,
	    int *n, int *k, int *m, int*p, int*s,
	    int *oa, int *da, double *cap,
	    double *gra, double *mcap, double *supp,
	    double *lowcb, double *upcb, int *acb,
	    int *kdb, int *knb, double *coefcb, double *flow,
	    void (f)(int*,int*,double*,double*,double*,int*));

void call_pprn(problem_type *p) {
#ifdef PRINTS
  printf("calling pprn2...\n");
#endif
  pprn2_(
	 p->ln
	 ,p->name
	 ,p->exit
	 ,p->n
	 ,p->k
	 ,p->m
	 ,p->p
	 ,p->s
	 ,p->oa
	 ,p->da
	 ,p->cap
	 ,p->gra
	 ,p->mcap
	 ,p->supp
	 ,p->lowcb
	 ,p->upcb
	 ,p->acb
	 ,p->kcb
	 ,p->ncb
	 ,p->coefcb
	 ,p->flow
	 ,NULL);
#ifdef PRINTS
  printf("returned from pprn2...\n");
#endif
}

void prob_file(problem_type *p) {
  int i,k;
  int M,N,K,P,S;
  FILE *fp;

#ifdef PRINTS
printf("enter prob_file %s\n",p->name);
#endif
  if (NULL==(fp=fopen(p->name,"w"))) {
    perror(p->name);
    return;
  }
  M=p->m[0];
  N=p->n[0];
  K=p->k[0];
  P=p->p[0];
  S=p->s[0];

  fprintf(fp,"%2d %2d %2d %2d %2d\n",M,N,K,P,S);
  fprintf(fp,"arc costs by commodities\n");
  for (k=0;k<K;k++) {
    for (i=0;i<N;i++) {
      fprintf(fp,"%7.1f ",p->gra[1+i+((N+1)*k)]);
    }
    fprintf(fp,"\n");
  }
  fprintf(fp,"arc capacities by commodities\n");
  for (k=0;k<K;k++) {
    for (i=0;i<N;i++) {
      fprintf(fp,"%7.1f ",p->cap[1+i+((N+1)*k)]);
    }
    fprintf(fp,"\n");
  }
  fprintf(fp,"node injections by commodities\n");
  for (k=0;k<K;k++) {
    for (i=0;i<M;i++) {
      fprintf(fp,"%7.1f ",p->supp[i+((M+1)*k)]);
    }
    fprintf(fp,"\n");
  }
  fprintf(fp,"arc mutual capacities\n");
  for (i=0;i<N;i++) {
    fprintf(fp,"%7.1f ",p->mcap[i]);
  }
  fprintf(fp,"\n");
  fprintf(fp,"network topology\n");
  for (i=0;i<N;i++)
    fprintf(fp,"%2d %2d\n",p->oa[i],p->da[i]);
  fprintf(fp,"lower and upper bounds\n");
  for (i=0;i<P;i++)
    fprintf(fp,"%7.1f ",p->lowcb[i]);
  fprintf(fp,"\n");
  for (i=0;i<P;i++)
    fprintf(fp,"%7.1f ",p->upcb[i]);
  fprintf(fp,"\n");
  fprintf(fp,"side constraints \n");
  for (i=0;i<S;i++)
    fprintf(fp,"%2d %2d %2d %7.1f\n",p->acb[i],p->kcb[i],p->ncb[i],p->coefcb[i]);
  fclose(fp);
#ifdef PRINTS
  printf("exit prob_file\n");
#endif
}



#include "parse_castro.h"

problem_type *the_problem;
int save_file_after_parse = 0;

REAL multi(body *bodies, int bodycount,
	   joint *joints, int jointcount,
	   struct appliedforce *forces,
	   int forcecount,
	   problem_type **pp) {

  static problem_type *p = NULL;
  /*  static itercount = 0; */

#ifdef PRINTS
  printf("entering multi\n");
#endif
  /*  printf("multi %d\n",++itercount);********/

  if (p == NULL) {
    the_problem = p = malloc(sizeof *p);
    if (p==NULL) {
      perror("can't alloc p");
      exit(77);
    }
  }

  *pp = p;
  parse_castro( p,
	 bodies,bodycount,joints,jointcount,forces,forcecount);

  if (save_file_after_parse) {
    prob_file(p);
  }

  /*
  if (itercount == 10) {
    mmbricks();
    prob_file(p);
    }

  for (i=0;i<((p->n[0]+1)*(p->k[0]));i++) p->flow[i]=-1;
  */
  call_pprn(p);

  if (*p->exit < 0) {
    printf("ERROR: pprn exit status %d\n",*p->exit);
    return 0.0;
  }


  else if (*p->exit == 0.0) {
#ifdef PRINTS
    printf("multi return 1.0\n");
#endif
    return 1.0;
  }
#ifdef PRINTS
  printf("multi returning 0.0\n");
#endif

  return 0.0;
}

#ifdef PARSE3
#include PARSE_INCLUDE



double itermultitest(body *bodies, int bodycount,
		   joint *joints, int jointcount,
		   struct appliedforce *forces,
		   int forcecount,
		   problem_type **pp) {

  static problem_type *p = NULL;
  /*  static itercount = 0; */
  double flow;

#ifdef PRINTS
  printf("enter itermultitest (one pass v)\n");
#endif
  /*  printf("multi %d\n",++itercount);********/

  if (p == NULL) {
    the_problem = p = malloc(sizeof *p);
    if (p==NULL) {
      perror("can't alloc p");
      exit(77);
    }
  }

  *pp = p;
  PARSE_ROUTINE( p,
	 bodies,bodycount,joints,jointcount,forces,forcecount);

  if (save_file_after_parse) {
    prob_file(p);
  }
  call_pprn(p);

  if (*p->exit != 0) {
    printf("-1 ERROR: pprn exit status %d\n",*p->exit);
    return 0.0;
  }

  flow=(p->flow[p->n[0]+((p->k[0]-1)*(p->n[0]+1))]);

#ifdef PRINTS
  printf("itermultitest3 returning %f\n",flow);
#endif
  return flow;

}
#else
double itermultitest(body *bodies, int bodycount,
		     joint *joints, int jointcount,
		     struct appliedforce *forcelist,
		     int forcecount,
		     problem_type **pp) {
  static double tests[]={0.0, 0.25, 0.5, 0.75, 0.875, .9375, 1.0, 1.0};
  force2d lastforce;
  int testcount = 8;
  double flow;
  int lower = 0; /* point where we know it succeeds */
  int upper = testcount-1; /* where we know it succeeds, notice the trick */
  int current;
  int mysave=0;


  lastforce=forcelist[forcecount-1].f;


  if (save_file_after_parse) {
    mysave=1;
    save_file_after_parse = 0;
  }

  while ((lower+1) != upper) {
    current = (upper + lower) / 2;
    forcelist[forcecount-1].f = mul2d(tests[current],lastforce);
    flow=multi(bodies, bodycount ,
	       joints, jointcount,
	       forcelist,forcecount,pp);
    if (flow == 0.0) {
      upper = current;
    }
    else
      lower = current;
  }
  forcelist[forcecount-1].f = lastforce;

#ifdef PRINTS
  printf("itermultitest returns %f\n",((tests[lower]+tests[upper])/2.0));
#endif
  if (mysave) {
    multi(bodies, bodycount ,
	  joints, jointcount,
	  forcelist,forcecount,pp);
    save_file_after_parse = 1;
  }
  return ((tests[lower]+tests[upper])/2.0);
}
#endif
