/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#include "headers.h"

static char bricksizes[] = { BRICKSIZES_LIST };

int bricksize(int x) {
  if (x<0) return bricksizes[0];
  if (x>=BRICKSIZES) return bricksizes[BRICKSIZES-1];
  return bricksizes[x];
}

int randombricksize() {
  return bricksize(RANDINT() % BRICKSIZES);
}



int find_parent_tree(int where,brick *bricks,int brickno, int *jointno) {
  int i,j;
  int r;

  for (i=0;i<BRICKJOINTS;i++) {
    j=bricks[where].joints[i];
    if (j == brickno) {
      *jointno = i;
      return where;
    }
    else if (j>=0) {
      r=find_parent_tree(j,bricks,brickno,jointno);
      if (r>=0) return r;
    }
  }
  return -1;
}

int find_parent(xybrickptr *roots, int rootcount,
		brick *bricks, int brickno, int *jointno) {
  int i;
  int r;
  for (r=0;r<rootcount;r++) {
    if (roots[r].no == brickno) {
      /* isa root */
      *jointno=-1;
      return r;
    }
    else {
      i=find_parent_tree(roots[r].no,bricks,brickno,jointno);
      if (i>=0) return i; /* found daddy */
    }
  }
  /* parent not found, orphan brick */
  abortonerror(538,"orphan brick %d on find_parent",brickno);
  return 0; /* unreachable, just to avid a warning */
}

int kill_brick(xybrickptr *myroots,
	       int *myrootcount,
	       brick *bricks,
	       int brickcount,
	       int brickno) {
  int myparent;
  int jointno;
  int i,j;
#ifdef DEBUGMUTATE
  printf("kill_brick %d \n",brickno);
#endif
  if (brickcount == 1) return 0;
  /* find a descendant of brickno that's a leaf */
  do {
    /* count descendants */
    i=0;
    for(jointno=0;jointno<BRICKJOINTS;jointno++)
      if (bricks[brickno].joints[jointno]!=EMPTY) i++;
    /* if more than one, choose a random descendant */
    if (i>0) {
      j=RANDINT()%i;
      /* find the j-th descendant */
      for (jointno=0;jointno<BRICKJOINTS;jointno++)
	if (bricks[brickno].joints[jointno]!=EMPTY) {
	  i++;
	  if (i>j) break;
	}
      brickno=bricks[brickno].joints[jointno];
    }
  } while (i>0);

  myparent =
    find_parent(myroots,*myrootcount,bricks,brickno,&jointno);


  if (jointno<0) {
    /* isa root */
    for (i=myparent;i<*myrootcount;i++) {
      myroots[i]=myroots[i+1];
    }
    (*myrootcount)--;
  }
  else {
    bricks[myparent].joints[jointno]=EMPTY;
    bricks[myparent].jointpos[jointno].x=
      bricks[myparent].jointpos[jointno].y=0;
  }
  return 0;
}

int shift_brick(xybrickptr *myroots,
		int myrootcount,
		brick *bricks,int brickno, int lr,int push) {
  /* tryto move brick brickno to the left/right
     lr==1 move to the right, -1 move to the left
     push 1/0 =  try/don't try to push descendants with me
  */
  int myparent;
  int jointno;
  int i,j;
#ifdef DEBUGMUTATE
  printf("shift_brick %d %d %d\n",brickno,lr,push);
#endif

  myparent = find_parent(myroots,myrootcount,bricks,brickno,&jointno);

  if (jointno<0) {
    /* isa root */
    myroots[myparent].x += lr;
  }
  else {
    bricks[myparent].jointpos[jointno].x += lr;
  }

  if (!push) {
    /* correct joints positions */
    for (i=0;i<4;i++) {
      j=bricks[brickno].joints[i];
      if (j>=0) {
	bricks[brickno].jointpos[i].x -= lr;
      }
    }
  }
  return 0;
}



int find_size(int size) {
  int i;
  for (i=0;i<BRICKSIZES;i++) {
    if (size == bricksizes[i]) return i;
  }
  printf("error: illegal size %d\n",size);
  abortonerror(212,"find_size");
  return 0;
}

int shrink_brick(xybrickptr *myroots, int myrootcount,
		 brick *bricks, int brickno, int lr) {
  int size;
  int newsize;
  int myparent,jointno;
  int i,j;


#ifdef DEBUGMUTATE
  printf("shrink_brick %d %d\n",brickno,lr);
#endif

  size=bricks[brickno].size;
  i = find_size(size);
  if (i==0) return 1; /* can't shrink */
  newsize=bricksizes[i-1];


  if (lr == 1) {
    /* shrink from the right */
    bricks[brickno].size = newsize;
  }
  else {
    /* shrink from the left */
    myparent=find_parent(myroots,myrootcount,bricks,brickno,&jointno);
    if (jointno < 0) {
      /* isa root */
      myroots[myparent].x += (size - newsize);
    }
    else {
      bricks[myparent].jointpos[jointno].x += (size - newsize);
    }
    bricks[brickno].size=newsize;
    /* adjust joints */
    for (i=0;i<4;i++) {
      j=bricks[brickno].joints[i];
      if (j>=0) {
	bricks[brickno].jointpos[i].x -= (size - newsize);
      }
    }
  }
  return 0;
}

int grow_brick(xybrickptr *myroots, int myrootcount,
	       brick *bricks, int brickno, int lr) {
  int size;
  int newsize;
  int i,j;
  int myparent,jointno;

#ifdef DEBUGMUTATE
  printf("grow_brick %d %d\n",brickno,lr);
#endif

  size = bricks [brickno].size;

  i=find_size(size);
  if (i==(BRICKSIZES-1)) return 1; /* can't grow */

  newsize=bricksizes[i+1];


  if (lr == 1) {
    /* grow on the right */
    bricks[brickno].size = newsize;
  }
  else {
    /* grow on the left */
    myparent=find_parent(myroots,myrootcount,bricks,brickno,&jointno);
    if (jointno < 0) {
      /* parent isa root */
      myroots[myparent].x -= (newsize - size);
    }
    else {
      /* parent isa brick */
      bricks[myparent].jointpos[jointno].x -= (newsize - size);
    }
    bricks[brickno].size = newsize;
    /* adjust joints */
    for (i=0;i<4;i++) {
      j=bricks[brickno].joints[i];
      if (j >= 0) {
	bricks[brickno].jointpos[i].x += (newsize - size);
      }
    }
  }
  return 0;
}

int adopt_brick(int son,int father,int newfather,int recurse) {
  int i;
  int whichjoint;

  if (father == GROUND) {
    /* unroot brick */
    for (i=0;i<rootcount;i++)
      if (roots[i].no == son) break;
    if (roots[i].no != son) abortonerror(757,"wrong root adoption");
    while (i < rootcount-1) {
      roots[i]=roots[i+1];
      i++;
    }
    rootcount--;
  }
  else {
    /* disconnect from father */
    for (i=0;i<BRICKJOINTS;i++)
      if (bricks MINE [father].joints[i] == son) break;

    if (i == BRICKJOINTS) {
      printf("jump_brick: %d is not %d's dad!!\n",father,son);
      abortonerror(221,"jump brick");
    }
    bricks MINE [father].joints[i] = EMPTY;
    bricks MINE [father].jointpos[i].x =
      bricks MINE [father].jointpos[i].y = 0;
  }

  for (i=0;i<BRICKJOINTS;i++)
    if (bricks MINE [newfather].joints[i] == EMPTY)
      break;
  whichjoint=i;

  if (newfather == GROUND) {
    /* root brick */
    if (rootcount == MAXROOTS) return 1; /* no empty roots */
    roots[rootcount].x = brickpos[son].x;
    roots[rootcount].y = brickpos[son].y;
    roots[rootcount].no = son;
    rootcount++;
  }
  else {
    /* adopt brick by newfather */
    if ((whichjoint = i) == BRICKJOINTS) return 1; /* no empty joints */
    bricks MINE [newfather].joints[whichjoint] = son;
    bricks MINE [newfather].jointpos[whichjoint].x =
      brickpos [son].x - brickpos[newfather].x;
    bricks MINE [newfather].jointpos[whichjoint].y =
      brickpos [son].y - brickpos[newfather].y;
  }
  parent[son]=newfather;

  if (recurse &&  GROUND != father) {
      /* try to adopt father, recursively! */
      adopt_brick(father,parent[father],son,1);
      }
  return 0;
  }

int jump_brick(xybrickptr *mutroots,int *mutrootcount,
	       brick *mutbricks,int *mutbrickcount) {
  int which;
  int father,newfather,son;

#ifdef DEBUGMUTATE
  printf("jump_brick\n");
#endif

  loadmodel(mutroots,*mutrootcount,mutbricks);
  makemodel();
  while((adoptcount > 0) &&
	((adoptives[adoptcount-1].son) >= (brickcount - PREDEF_BRICK_NO))) {
    adoptcount--;
  }
  if (adoptcount MINE <= 0) return 1; /* no jumps available */
  which = RANDINT() % adoptcount MINE;

  son = adoptives MINE [which].son;
  newfather = adoptives MINE [which].father;
  father = parent MINE [son];

  adopt_brick(son,father,newfather,0);
  ARRCPY(mutroots,roots,rootcount);
  ARRCPY(mutbricks,bricks,brickcount);
  *mutrootcount=rootcount;
  *mutbrickcount=brickcount;
  return 0;
}


int change_flow(xybrickptr *mutroots, int mutrootcount,
		brick *mutbricks,int mutbrickcount) {
  int which;
  int son;
  int i;

#ifdef DEBUGMUTATE
  printf("change_flow\n");
#endif

  loadmodel(mutroots, mutrootcount,mutbricks);
  makemodel();
  while((adoptcount > 0) &&
	((adoptives[adoptcount-1].son) >= (brickcount - PREDEF_BRICK_NO))) {
    adoptcount--;
  }
  if (adoptcount MINE <= 0) return 1; /* no jumps available */
  which = RANDINT() % adoptcount MINE;
  son = adoptives MINE [which].son;
  for (i=0;i<BODYJOINTS;i++)  /* mutate weight distribution  */
    bricks MINE[son].flow[i] *= (1.0 +
				 ((REAL)((RANDINT()%3000)-1500))/5000.0 );
  cmemcpy(mutbricks,bricks MINE,brickcount,sizeof *mutbricks);
  return 0;
}


int change_order(xybrickptr *myroots,int myrootcount,
		 brick *mybricks, int mybrickcount) {
  int multi[MAXBRICKS+1];
  int multicount;
  int i,j,k;

#ifdef DEBUGMUTATE
  printf("change_order\n");
#endif
  multicount=0;
  for (i=0;i<mybrickcount;i++) {
    k=0;
    for (j=0;j<BRICKJOINTS;j++)
      if (mybricks[i].joints[j] >=0) k++;
    if (k>1) {
      multi[multicount]=i;
      multicount++;
    }
  }
  if (myrootcount > 1) {
    multi[multicount]=-1;
    multicount++;
  }
  if (multicount == 0) return 1;
  k=RANDINT() % multicount;
  if (multi[k] == -1) {
    /* switch roots */
    xybrickptr aroot;
    i=RANDINT() % myrootcount;
    j=RANDINT() % (myrootcount-1);
    if (j>=i) j++;
    aroot=myroots[i];
    myroots[i]=myroots[j];
    myroots[j]=aroot;
  }
  else {
    /* switch joints */
    int j1,j2;
    int joi;
    jointpos_type pos;
    i=multi[k]; /* i is brickno now */
    j1=j2=-1;
    for (k=0,j=0;j<BRICKJOINTS;j++)
      if (mybricks[i].joints[j] >= 0) {
	k++;
	if (k==1) j1=j;
	else if (k==2) j2=j;
	else switch(RANDINT()%3) {
	case 0: j1=j; break;
	case 1: j2=j; break;
	}
      }
#ifdef CHECKS
    if (j1<0 || j2<0) abortonerror(453,"switch joints brick %d",i);
#endif
    joi=mybricks[i].joints[j1];
    pos=mybricks[i].jointpos[j1];
    mybricks[i].joints[j1]=mybricks[i].joints[j2];
    mybricks[i].jointpos[j1]=mybricks[i].jointpos[j2];
    mybricks[i].joints[j2]=joi;
    mybricks[i].jointpos[j2]=pos;
  }
  return 0;
}

int add_son_or_root(xybrickptr *myroots, int *myrootcount,
		    brick *mybricks, int *mybrickcount) {
  int which;
  short i,dx,dy,bx,by,bs,size,joint=0,ground,ux,lowx,hix;
  int ground_count=object_kind_count[GROUND_KIND];
  obj_type *grounds=object_kind[GROUND_KIND];

  if (*mybrickcount >= MAXBRICKS)
    return 1;

  loadmodel(myroots,*myrootcount,mybricks);
  makemodel();

  which=RANDINT() % (brickcount + ground_count);
  if (which < brickcount) {
    /* add a brick's descendant */
    for (i=0;i<4;i++) {
      if (bricks[which].joints[i] == EMPTY) break;
    }
    if (i==4) return 1;
    joint=i;
    bx=brickpos[which].x;
    by=brickpos[which].y;
    bs=bricks[which].size;
  }
  else {
    /* add a ground's descendant */
    if ((rootcount+1)>=MAXROOTS)
      return 1;
    ground=which-brickcount;
    bx=grounds[ground].x;
    by=grounds[ground].y;
    bs=grounds[ground].size;
  }

  dy=(RANDINT()%2)*2-1; /* +-1 */
  by=dy+by;
  if (by < MINY || by > MAXY) return 1;
  i=dx=RANDINT() % bs;
  ux=(RANDINT()%2)*2-1;
  /* try to find an anchor position */
  while(dx>=0 && dx<bs && getgrid(bx+dx,by) != EMPTY)
    dx+=ux;
  if ((dx<0) || (dx>=bs)) {
    ux=(-ux);
    dx=i+ux;
    while(dx>=0 && dx<bs && getgrid(bx+dx,by) != EMPTY)
      dx+=ux;
  }
  if ((dx<0) || (dx>=bs)) return 1; /* failed */
  for (lowx=bx+dx-1;
       lowx>=minx && getgrid(lowx,by) == EMPTY;lowx--);
  lowx++;
  for (hix=bx+dx+1;
       hix<=maxx && getgrid(hix,by) == EMPTY;hix++);
  if ((hix-lowx) < bricksizes[0]) return 1; /* can't fit a brick in there */
  size=RANDINT()%BRICKSIZES;
  while (bricksizes[size] > (hix -lowx)) size--;
  size=bricksizes[size];
  dx -= size/2;
  if (lowx > bx+dx) dx=(lowx-bx);
  if (hix < bx+dx+size) dx=(hix-size-bx);
  bricks[brickcount]=newbrick(size);
  if (which < brickcount) {
    bricks[which].joints[joint]=brickcount;
    bricks[which].jointpos[joint].x=dx;
    bricks[which].jointpos[joint].y=dy;
  }
  else {
    roots[rootcount].x=bx + dx;
    roots[rootcount].y=by;
    roots[rootcount].no=brickcount;
    rootcount++;
  }
  brickcount++;
  ARRCPY(mybricks,bricks,brickcount);
  *mybrickcount=brickcount;
  ARRCPY(myroots,roots,rootcount);
  *myrootcount=rootcount;
  return 0;
}


int identical(xybrickptr *r1,int rc1,brick *b1,int bc1,
	      xybrickptr *r2,int rc2, brick *b2, int bc2) {
  int i,j;

  if (rc1 != rc2) return 0;
  if (bc1 != bc2) return 0;
  for (i=bc1-1;i>=0;i--) {
    if (b1[i].size != b2[i].size) return 0;
    for (j=0;j<BRICKJOINTS;j++)
      if (b1[i].joints[j] != b2[i].joints[j] ||
	  b1[i].jointpos[j].x != b2[i].jointpos[j].x ||
	  b1[i].jointpos[j].y != b2[i].jointpos[j].y) return 0;
    for (j=0;j<BODYJOINTS;j++)
      if (b1[i].flow[j] != b2[i].flow[j]) return 0;
  }
  for (i=0;i<rc1;i++) {
    if (r1[i].x != r2[i].x) return 0;
    if (r1[i].y != r2[i].y) return 0;
    if (r1[i].no != r2[i].no) return 0;
  }

  return 1;
}

#ifdef DEBUGMUTATE
#define HANDMUTATE
#endif

int mutate(xybrickptr *someroots, int somerootcount,
	   brick *somebricks, int somecount) {
  USEMINE
    int what;
  int ret;
  int result;
  int heavy;
  int lr,push,brickno;
  int arootcount;
  LOCAL brick atree[MAXBRICKS];
  LOCAL xybrickptr aroots[MAXROOTS];
  int acount;
#ifdef HANDMUTATE
  static brick atree0[MAXBRICKS];
  static int acount0;
  static xybrickptr aroots0[MAXROOTS];
  static int arootcount0;
#endif

  acount=somecount;
  ARRCPY(aroots,someroots,somerootcount);
  arootcount=somerootcount;
  ARRCPY(atree,somebricks,acount);
  acount=somecount;

#ifdef HANDMUTATE
	   loop:
  memcpy(atree0,atree,sizeof atree0);
  memcpy(aroots0,aroots,sizeof aroots);
  arootcount0=arootcount;
  acount0=acount;
#endif
#ifdef DEBUGMUTATE
  printf("mutate saves PARENT.m\n");
  loadmodel(aroots,arootcount,atree);
  makemodel();
  tooheavy();
  mat_bricks("PARENT.m");
#endif

#ifdef PRINTS
  printf("MUTATE s=%d",acount);
#endif

  /* mutations:
     0: shift (l/r w/wo carry)
     1: shrink (l/r w/wo carry)
     2: grow (l/r w/wo carry)
     3: create_son
     4: jump (from one branch to another, if possible)
     5 : redistribute weight
     6 : reorder
     disabled : kill brick
  */

  what=RANDINT() % 7;

  /* one joint of i-th brick will be mutated, or
     the root if -1 */

  result=-1;
  switch(what) {
  case 0:
    /* shift */
    if (acount <= 0) break;
    brickno = RANDINT() % acount;
    lr=2*(RANDINT()%2)-1;
    push=RANDINT()%2;
    result=shift_brick(aroots,arootcount,atree,brickno,lr,push);
    break;

  case 1:
    /* shrink */
    if (acount <= 0) break;
    brickno=RANDINT() % acount;
    if (brickno == 0) lr=1;
    else
      lr=(2*(RANDINT()%2))-1;
    result=shrink_brick(aroots,arootcount,atree,brickno,lr);
    break;

  case 2:
    /* grow */
    if (acount <= 0) break;
    brickno=RANDINT() % acount;
    lr=(2*(RANDINT()%2))-1;
    result=grow_brick(aroots,arootcount,atree,brickno,lr);
    break;

  case 3:
    /*
    brickno=RANDINT() % acount;
    result = add_son(atree,brickno,acount);
    if (result == 0) acount++;*/
    result=add_son_or_root(aroots, &arootcount,atree,&acount);
    break;

  case 4:
    result = jump_brick(aroots,&arootcount,atree,&acount);
    break;

  case 5:
    result=change_flow(aroots,arootcount,atree,acount);
    break;

  case 6:
    result=change_order(aroots,arootcount,atree,acount);
    break;

  case 998:
    brickno=RANDINT() % acount;
    result=kill_brick(aroots,&arootcount,atree,acount,brickno);
    break;
  }

#ifdef DEBUGMUTATE
  printf("mutation result=%d\n",result);
#endif

  ret=0;

  if (result != 0)
    ret=1;
  else if (loadmodel(aroots,arootcount,atree)) {
#ifdef DEBUGMUTATE
    printf("load failed\n");
#endif
    ret= 1;
  }
  else if (identical(roots,rootcount,bricks,brickcount,
		     someroots,somerootcount,somebricks,somecount)) {
#ifdef DEBUGMUTATE
      printf("mutation failed: Identical\n");
#endif
      ret= 1;
  }

  else if (makemodel()) {
    printf("Error: Makemodel failed at mutate\n");
    /* exit(1); */ ret= 1; /* mutation failed */
  }
  if (ret==0) {
    heavy=tooheavy();
#ifdef DEBUGMUTATE
    printf("MUTATION DONE RESULTS in MUTANT.m\n");
    mat_bricks("MUTANT.m");
    if (!heavy) printf("fitness=%f\n",fitnessvalue = fitness());
    else printf("Is heavy\n");
#endif
  }
#ifdef HANDMUTATE
  {
    char s[200];
    printf("keep or discard: ");
    scanf("%s",s);
    if (s[0]=='d') {
      ARRCPY(aroots,aroots0,MAXROOTS);
      ARRCPY(atree,atree0,MAXBRICKS);
      acount=acount0;
      arootcount=arootcount0;
      goto loop;
    }
    else {
      memcpy(aroots,roots,sizeof aroots);
      memcpy(atree,bricks, sizeof atree);
      acount=brickcount;
      arootcount=rootcount;
      goto loop;
    }
  }
#endif
  if (ret==0) {
    if (heavy) ret= 1;   /* collapsed! */
    else  fitnessvalue MINE= fitness();
  }
    return ret;
}



#define loadtree( gx, gy, a) loadxtree((gx),(gy),(a),UNUSED,0,NULL,0,minx,maxx-minx)


int loadxtree(int gx,int gy,brick *a, int acut, int aoff,
	     brick *b, int boff,
	     int px, int psize) {
  /*
     reads brick tree from array a and loads them
     crosses over replacing a[acut] with b[boff]
     returns 1 if at least one brick was added
     px,psize: parent x coord and size
  */

  USEMINE
  int j,k,joint,i;
  if (brickcount MINE == (MAXBRICKS-PREDEF_BRICK_NO)) return 1; /* too big; failed */
  if ((gx+a[0].size <= px) || (px+psize <= gx))
    return 1; /* improper position */
  if (putbrickat(gx,gy,a[0].size,brickcount))
    return 1; /* doesn't fit */
  k=brickcount MINE++;
  bricks MINE [k].size = a[0].size;
  brickpos MINE [k].x = gx;
  brickpos MINE [k].y = gy;
  OBJCPY(bricks MINE [k].flow,a[0].flow);
  for (j=0;j<4;j++) {
    joint = a[0].joints[j];
    if (!ISBRICK(joint)) {
      bricks MINE [k].joints[j] = joint;
      bricks MINE [k].jointpos[j] = a[0].jointpos[j];
    }
    else /* if (joint >= 0) */ {
      bricks MINE [k].joints[j] = brickcount MINE ;
      bricks MINE [k].jointpos[j] = a[0].jointpos[j];
      if (joint == acut)
	i = loadxtree(gx+a[0].jointpos[j].x,
		     gy+a[0].jointpos[j].y,
		     b,UNUSED,boff,NULL,0,gx,a[0].size);
      else
	i = loadxtree(gx+a[0].jointpos[j].x,
		     gy+a[0].jointpos[j].y,
		     &a[joint-aoff],acut,joint,b,boff,
		     gx,a[0].size);
      if (i) {
	/* didn't fit, cut off rest */
	bricks MINE[k].joints[j]=EMPTY;
	bricks MINE[k].jointpos[j].x=
	  bricks MINE[k].jointpos[j].y = 0;
      }
    }
  }
  return 0;
}



#define loadroot(gx, gy, a) loadxroot((gx),(gy),(a),UNUSED,0,NULL,0)

int loadxroot(int gx, int gy, brick *a,
	      int acut, int aoff,
	      brick *b, int boff) {
  /*
     loads subtree a at gx gy
     adds root to working mem
     crossover with b possible
  */
  int curbrick;
  int ret;
  if (rootcount == MAXROOTS) return 1;
  curbrick=brickcount;
  if (loadxtree(gx,gy,a,acut,aoff,b,boff,minx,maxx-minx)) {
    ret=1;
#ifdef CHECKS
    if (brickcount != curbrick) {
      abortonerror(878,"loadtree failed but no");
    }
#endif
  }
  else {
    roots[rootcount].x=gx;
    roots[rootcount].y=gy;
    roots[rootcount].no=curbrick;
    rootcount++;
    ret=0;
  }
  return ret;
}

#define loadall(ar,arcount,a) loadxall((ar),(arcount),(a),UNUSED,NULL,0)

int loadxall(xybrickptr aroots[], int arootcount,
	     brick *a, int acut,
	     brick *b, int boff) {
  /*
     loads multitree a
     introduces crossover with b[boff] at acut
  */
  int r;
  int ret;
  int what;

  if (arootcount == 0) return 0;
  ret=1;
  for (r=0;r<arootcount;r++) {
    if (aroots[r].no==acut)
      what=loadxroot(aroots[r].x,aroots[r].y,b,
		UNUSED,boff,NULL,0);
    else
      what=loadxroot(aroots[r].x,aroots[r].y,
		&a[aroots[r].no],
		acut,
		aroots[r].no,
		b,boff);
    if (what==0) ret=0;
  }
  return ret;
}

int loadmodel(xybrickptr *my_roots,
	      int my_rootcount, brick *my_bricks) {
  /* resets and loads all model */
  initgrid();
  brickcount MINE = bodycount MINE = 0;
  rootcount = 0;
  return loadall(my_roots,my_rootcount,my_bricks);

}


int crossover(xybrickptr ar[], int arcount,
	      brick a[], int acount,
	      xybrickptr br[], int brcount,
	      brick b[], int bcount) {
  USEMINE
    int i,j;
  int ret;
  short gx,gy;

#ifdef DEBUGMUTATE
  printf("crossover saves PARENT and PARENT2\n");
  loadmodel(ar,arcount,a);
  makemodel();
  tooheavy();
  mat_bricks("PARENT.m");
  loadmodel(br,brcount,b);
  makemodel();
  tooheavy();
  mat_bricks("PARENT2.m");
#endif
  /* no crossover when one parent is empty;
     should be replaced with creating a root
     pointing to a random substructure on the other
  */
  if (acount == 0 || bcount == 0) return 0;

  i = RANDINT() % acount;
  j = -1;
  if ( (RANDINT() % 2) == 0) {
    /* attempt a positional crossover */
    loadmodel(ar,arcount,a);
    gx = brickpos[i].x;
    gy = brickpos[i].y;
#ifdef CHECKS
    if (getgrid(gx,gy) != i) {
      fprintf(stderr,"error: brick %d is not at brickpos[i](%d,%d) but instead %d\n"
	      ,i,gx,gy,getgrid(gx,gy));
      abortonerror(342,"brick not at brickpos");
    }
#endif
    loadmodel(br,brcount, b);
    j=getgrid(gx,gy);
    if (j < 0) {
      /* there's no brick there, maybe we can try
	 a brickno relationship */
      if (i < brickcount) j=i;
    }
  }
  if (j < 0) {
    j = RANDINT() % bcount;
  }


#if defined( PRINTS) || defined(DEBUGMUTATE)
  printf("CROSS (%d)%d (%d)%d>",i,acount,j,bcount);
#endif

  initgrid();
  brickcount=0;
  rootcount=0;
  ret=0;
  if (loadxall(ar,arcount,a,i,&b[j],j)) {
#ifdef DEBUGMUTATE
    printf("loadxall failed\n");
#endif
    ret=1;
  }
  else if (identical(roots,rootcount,bricks,brickcount,ar,arcount,a,acount)) {
    ret=1;
#ifdef DEBUGMUTATE
    printf("Are identical\n");
#endif
  }
  else if (makemodel()) {
    ret=1;
#ifdef DEBUGMUTATE
    printf("makemodel failed\n");
#endif
  }
  else if (tooheavy()) {
#ifdef DEBUGMUTATE
    printf("too heavy\n");
#endif
    ret=1;
  }
  else fitnessvalue=fitness();
#ifdef DEBUGMUTATE
  printf("crossover returns %d\n",ret);
  mat_bricks("MUTANT.m");
  printf("mutant saved in MUTANT.m\n");
#endif

  return ret;
}
