#define NEWPARSE

void parse_castro(problem_type *p,
	   body *bodies, int bodycount,
	   joint *joints, int jointcount,
	   struct appliedforce *forces,
	   int forcecount) {
  int i,j,k;
  int head, tail;
  int arcno,arccount;
  double tau,a;
  joint *joint;
  int sign;
  int index;
  int sideno;
  int sidecount;
  static short sideconvert[MAX_SIDE];
  
  p->ln[0]=0;
  strcpy(p->name,"bogus.dat");
  p->m[0] = bodycount+3; /* bricks + 1 = no of nodes */
  /* nodes 1...bodycount are the bricks
     node bodycount+1 is the ground 
     node bodycount+2 is the sink 
     node bodycount+3 is the intermediate between those two */

  arccount = 0;
  /* count arcs: no of arcs = no of joints */
  for (i=0;i<bodycount;i++) 
    arccount += bodies[i].jointcount;

  arccount++; /* an extra arc to connect ground to sink */
  arccount++; /* and another */
  arccount += forcecount; /* return arcs go from sink to entry point of each force */
  
  p->n[0] = arccount; /* no. of arcs */
  p->k[0] = forcecount; /* no of commodities = forces */
  p->p[0] = jointcount; /* one side constraint per joint */
  p->s[0] = (arccount - 2 - forcecount)  * forcecount; /* one factor per (arc,commodity) pair  */
  
  if (p->m[0] >= MAX_PARAM_M 
      || p->n[0] >= MAX_PARAM_N
      || p->k[0] >= MAX_PARAM_K
      || p->p[0] >= MAX_PARAM_P
      || p->s[0] >= MAX_PARAM_S
      ) {
    printf("error: too big %d %d %d %d %d %d %d %d %d %d\n",
	   p->m[0],p->n[0],p->k[0],p->p[0],p->s[0],
	   MAX_PARAM_M,MAX_PARAM_N,MAX_PARAM_K,MAX_PARAM_P,MAX_PARAM_S);
    abortonerror(33,"too big");
  }

  for (i=0;i<=jointcount;i++) sideconvert[i]=0;

  /* compute network */
  /* 1. supplies, all zero */
  
  for (k=0;k<forcecount;k++) 
    for (i=0;i<p->m[0];i++) 
      p->supp[i+(k*(p->m[0]+1))] = 0.0;
      
  
  /* side constraints */
  for (j=0;j<jointcount;j++) {
    p->upcb[j] = joints[j].maxtorque;
    p->lowcb[j] = - p->upcb[j];
  }
      
  /* 2. topology, capacities */

  arcno = 0;
  sideno = 0;
  sidecount = 0;
  for (i=0;i<bodycount;i++) {
    for (j=0;j<bodies[i].jointcount;j++,arcno++) {
      head=i+1;
      if (bodies[i].joins[j] == NULL)
	tail =  bodycount+1;
      else
	tail = 1 + bodies[i].joins[j]->n;
      /* for every arc */
      if (head < 0) {
	printf("head = %d\n",head);
	abortonerror(99,"head");
      }
      p->oa[arcno] = head;
      p->da[arcno] = tail;
      p->mcap[arcno] = 1.0 * forcecount;
      
  
      
      /* calc capacities for each commodity */
      
      for (k=0;k<forcecount;k++) {
	joint = bodies[i].joints[j];
	sign = bodies[i].jointsign[j];
	
	tau = sign * torqueat(joint->x,forces[k].x,forces[k].f);
	a=fabs(tau);
	index = 1+arcno+(k*(arccount+1));

	/*
	if (a < joint->maxtorque) 
	  p->cap[index] = 1.0;
	else
	  p->cap[index] = joint->maxtorque / a; 
	*/
	p->cap[index] = 1.0; /* let only the side constraints handle capacities */
	p->gra[index] = 0.0; /* a/ joint->maxtorque */ 
	
	if (a != 0.0) {
	  /* side constraint */
	  p->acb[sideno] = 1+arcno;
	  p->kcb[sideno] = 1+k;
	  p->ncb[sideno] = 1+joint->n;
	  sideconvert[1+joint->n]=1;
	  p->coefcb[sideno] = tau;
	  sideno++;
	}
	else p->s[0]--;
      }
    }
  }

  /* delete empty side conditions */
  for (i=1,j=0;i<=jointcount;i++) {
    if(sideconvert[i]) sideconvert[i]=++j;
  }
  
  p->p[0] = j;

  for (i=0;i<sideno;i++) {
    p->ncb[i] = sideconvert[p->ncb[i]];
  }
  
  /* return arcs go from sink to commdities entries */

  for (k=0;k<forcecount;k++) {
    p->oa[arcno] = bodycount+2 ;
    p->da[arcno] = 1+forces[k].target;
    p->mcap[arcno] = 1.0;
    /* capacities are 0 except for the right thing */
    for (j=0;j<forcecount;j++) {
      index = 1+arcno+(j*(arccount+1));
      p->cap[index]=0.0;
      p->gra[index]=0.0;
    }
    index=1+arcno+(k*(arccount+1));
    p->cap[index]=1.0;
    p->gra[index]=0.0;
    arcno++;
  }
    
  /* this last arc connects ground(int) to sink */
  
  for (k=0;k<forcecount;k++) {
    index = 1+arcno+(k*(arccount+1));
    p->cap[index] = 1.0;
    p->gra[index] = -1.0;
  }

  p->oa[arcno]=bodycount+3;
  p->da[arcno]=bodycount+2;
  p->mcap[arcno]=1.0*forcecount;
  arcno++;
  
  /* repeated, ground to int  */
  for (k=0;k<forcecount;k++) {
    index = 1+arcno+(k*(arccount+1));
    p->cap[index] = 1.0;
    p->gra[index] = -1.0;
  }

  p->oa[arcno]=bodycount+1;
  p->da[arcno]=bodycount+3;
  p->mcap[arcno]=1.0*forcecount;
  arcno++;

  if (sideno != p->s[0]) {
    printf("wrong count sideno %d p->s[0] %d\n",
	   sideno,p->s[0]);
    abortonerror(37,"count sideno");
  }
  if (arcno != p->n[0]) {
    printf("wrong count arcno %d p->n[0] %d\n",
	   arcno,p->n[0]);
    abortonerror(38,"count arcno");
  }
}
