void parse_castro3(problem_type *p,
	   body *bodies, int bodycount,
	   joint *joints, int jointcount,
	   struct appliedforce *forces,
	   int forcecount) {
  int i,j,k;
  int head, tail;
  int arcno,arccount;
  double tau,a;
  joint *joint;
  int sign;
  int index;
  int sideno;
  int sidecount;
  static short sideconvert[MAX_SIDE+1];
  /* this version for calculating partial solutions */
  
  strcpy(p->name,"bogus.dat");
  p->ln[0]=0;
  p->m[0] = bodycount+2; /* bricks + 1 = no of nodes */
  /* nodes 1...bodycount are the bricks
     node bodycount+1 is the ground 
     node bodycount+2 is the sink */

  arccount = 0;
  /* count arcs: no of arcs = no of joints */
  for (i=0;i<bodycount;i++) 
    arccount += bodies[i].jointcount;

  arccount += forcecount; /* each force has a costly arc to divert */
  arccount++; /* an extra arc to connect ground to sink */
  
  p->n[0] = arccount; /* no. of arcs */
  p->k[0] = forcecount; /* no of commodities = forces */
  p->p[0] = jointcount; /* one side constraint per joint */
  p->s[0] = (arccount-1-forcecount) * forcecount; /* one factor per (arc,commodity) pair  */
  
  if (p->m[0] >= MAX_PARAM_M 
      || p->n[0] >= MAX_PARAM_N
      || p->k[0] >= MAX_PARAM_K
      || p->p[0] >= MAX_PARAM_P
      || p->s[0] >= MAX_PARAM_S
      ) {
    printf("error: too big %d %d %d %d %d %d %d %d %d %d\n",
	   p->m[0],p->n[0],p->k[0],p->p[0],p->s[0],
	   MAX_PARAM_M,MAX_PARAM_N,MAX_PARAM_K,MAX_PARAM_P,MAX_PARAM_S);
    exit(33);
  }

  for (i=0;i<=jointcount;i++) sideconvert[i]=0;

  /* compute network */
  /* 1. supplies */
  
  for (k=0;k<forcecount;k++) 
    for (i=0;i<p->m[0];i++) 
      p->supp[i+(k*(p->m[0]+1))] = 0.0;
      
  /* 2. all forces enter at their assigned entry brick 
        and exit at the sink */

  for (k=0;k<forcecount;k++) {
    p->supp[forces[k].target + (k*(p->m[0]+1))] = 1.0;
    p->supp[bodycount+1 + (k*(p->m[0]+1))] = -1.0;
  }
  
  /* side constraints */
  for (j=0;j<jointcount;j++) {
    p->upcb[j] = joints[j].maxtorque;
    p->lowcb[j] = - p->upcb[j];
  }
      
  /* 2. topology, capacities */

  arcno = 0;
  sideno = 0;
  sidecount = 0;
  for (i=0;i<bodycount;i++) {
    for (j=0;j<bodies[i].jointcount;j++,arcno++) {
      head=i+1;
      if (bodies[i].joins[j] == NULL)
	tail =  bodycount+1;
      else
	tail = 1 + bodies[i].joins[j]->n;
  
      /* for every arc */
      if (head < 0) {
	printf("head = %d\n",head);
	exit(99);
      }
      p->oa[arcno] = head;
      p->da[arcno] = tail;
      p->mcap[arcno] = 1.0 * forcecount;
      
      joint = bodies[i].joints[j];
      sign = bodies[i].jointsign[j];

      /* calc capacities for each commodity */   	  
      for (k=0;k<forcecount;k++) {
	tau = sign * torqueat(joint->x,forces[k].x,forces[k].f);
	a=fabs(tau);
	index = 1+arcno+(k*(arccount+1));

	/*
	if (a < joint->maxtorque) 
	  p->cap[index] = 1.0;
	else
	  p->cap[index] = joint->maxtorque / a; 
	*/
	p->cap[index] = 1.0; /* let only the side constraints handle capacities */
	p->gra[index] = 0.0; /* a / joint->maxtorque; */
	
	if (a != 0.0) {
	  /* side constraint */
	  p->acb[sideno] = 1+arcno;
	  p->kcb[sideno] = 1+k;
	  p->ncb[sideno] = joint->n;
	  sideconvert[joint->n]=1;
	  p->coefcb[sideno] = tau;
	  sideno++;
	}
	else p->s[0]--;
      }
    }
  }

  /* delete empty side conditions */
  for (i=1,j=0;i<=jointcount;i++) {
    if(sideconvert[i]) sideconvert[i]=++j;
  }
  
  p->p[0] = j;

  for (i=0;i<sideno;i++) {
    p->ncb[i] = sideconvert[p->ncb[i]];
  }
  
  /* diverting arcs connect each force straight to the ground, with cost */
  for (j=0;j<forcecount;j++) {
    for (k=0;k<forcecount;k++) {
      index = 1+arcno+(k*(arccount+1));
      if (j==k) {
	p->cap[index]=1.0;
	p->gra[index]=1.0;
      }
      else {
	p->cap[index] = 0.0;
	p->gra[index] = 0.0;
      }
    }
    p->oa[arcno]=1+forces[j].target;
    p->da[arcno]=bodycount+2;
    p->mcap[arcno]=1.0;
    arcno++;
  }
  

  /* final arc connects ground to sink to collect information */
  for (k=0;k<forcecount;k++) {
    index = 1+arcno+(k*(arccount+1));
    p->cap[index] = 1.0;
    p->gra[index] = 0.0;
  }

  p->oa[arcno]=bodycount+1;
  p->da[arcno]=bodycount+2;
  p->mcap[arcno]=1.0*forcecount;
  arcno++;

  if (sideno != p->s[0]) {
    printf("wrong count sideno %d p->s[0] %d\n",
	   sideno,p->s[0]);
    exit(37);
  }
  if (arcno != p->n[0]) {
    printf("wrong count arcno %d p->n[0] %d\n",
	   arcno,p->n[0]);
    exit(38);
  }
}

void unparse_3(problem_type *p,
	       body *bodies, int bodycount,
	       joint *joints, int jointcount,
	       struct appliedforce *forces,
	       int forcecount) {

  int i,j,k;
  int arcno;
  joint *joint;
  int sign;
  int index;
  int arccount;
  double tau;

  /* initialize all torques to zero */

  for (i=0;i< jointcount; i++) joints[i].torques = 0.0;
  
  /* go over the joints */
  
  arcno=0;
  arccount=p->n[0];

  for (i=0;i<bodycount;i++) {
    for (j=0;j<bodies[i].jointcount;j++,arcno++) {
      /* foud a joint, add flow for each commodity */
      
      for (k=0;k<forcecount;k++) {
	joint = bodies[i].joints[j];
	sign = bodies[i].jointsign[j];
	
	tau = sign * torqueat(joint->x,forces[k].x,forces[k].f);
	index = 1+arcno+(k*(arccount+1));
	joint->torques += tau * p->flow[index];
      }
    }
  }
}
      
  
