/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#define PROBLEM_C

#include "headers.h"

void reserve_spots(int mode) {
  int i,j;
  for (i=0;i<obj_count;i++) {
    if (objects[i].kind == FIXED_KIND)
      for (j=0;j<objects[i].size;j++)
	setgrid(i,j,mode);
  }
}

int add_predefs(int count) {
  static int first = 1;
  static brick *predef_bricks;
  static struct {short x,y;} *predef_pos;
  static int predef_count;
  int i,k;

  if (first) {
    first = 0;
    predef_count = 0;
    for (k=0;k<obj_count;k++)
      if (objects[k].kind == FIXED_KIND) predef_count++;
    ARRALLOC(predef_bricks,predef_count);
    ARRALLOC(predef_pos,predef_count);
    i=0;
    for (k=0;k<obj_count;k++)
      if (objects[k].kind == FIXED_KIND) {
	predef_pos[i].x = objects[k].x;
	predef_pos[i].y = objects[k].y;
	predef_bricks[i] = newbrick(objects[k].size);
      }
  }

  for (i=0;i<predef_count;i++) {
    bricks[count] = predef_bricks[i];
    if (addbrick(predef_pos[i].x, predef_pos[i].y,count,-1)) {
      printf("Error adding predefs\n");
      /* exit(1); */ return 1;
    }
    count++;
  }
  brickcount = count;
  return 0;
}


REAL netvalue;

int tooheavy() {
  USEMINE
  /* init joints and bodies */
  int i;
  problem_type *p;

  /* build force list */
  for (i=0;i<bodycount MINE;i++) {
    forcelist[i].target = i;
    forcelist[i].x = bodies MINE[i].x;
    forcelist[i].f = mul2d(bodies MINE[i].mass,grav2d);
  }

  forcecount=bodycount;
  if (simulation_mode == EMBEDDED_MODE)
    netvalue = flow(bodies,bodycount,joints,jointcount,
		    forcelist,forcecount);
  else
    netvalue=multi(bodies,bodycount,joints,jointcount,forcelist,bodycount,&p);

  if (netvalue < 0.9999) {
#ifdef PRINTS
    printf("Too heavy.\n");
#endif
    return 1;
  }
  else return 0;
}

REAL fitness() {
  int i,k,dx,dy,d;
  static int target_count=0;
  static double *target_dist;
  static int load_count=0;
  static struct {short x,y; } *targets;
  static struct {short x,y; REAL load;} *loads;
  static int first = 1;
  int alltargets;
  REAL fit;

  /* initialization */

  if (bodycount == 0) return 0;

  if (first) {

    for (k=0;k<obj_count;k++)
      if (objects[k].kind == TARGET_KIND
          || objects[k].kind == LOAD_KIND) {
        target_count++;
        if (objects[k].kind == LOAD_KIND)
          load_count++;
      }
    ARRALLOC(target_dist,target_count);
    ARRALLOC(targets, target_count);
    ARRALLOC(loads,load_count);

    for (k=0,i=0;k<obj_count;k++)
      if (objects[k].kind == TARGET_KIND
          || objects[k].kind == LOAD_KIND) {
        targets[i].x = objects[k].x;
        targets[i].y = objects[k].y;
        i++;
      }

    for (k=0,i=0;k<obj_count;k++)
      if (objects[k].kind == LOAD_KIND) {
        loads[i].x = objects[k].x;
        loads[i].y = objects[k].y;
        loads[i].load = ((REAL) objects[k].size) / WEIGHT;
        i++;
      }
    first = 0;
  }

  /* evaluate target point fitnesses */
  fit = 1.0;

  for (i=0;i<target_count;i++)
    target_dist[i]= 1000000000.0;

  for (i=0;i<brickcount;i++)
    for (k=0;k<target_count;k++) {
      if (brickpos[i].x > targets[k].x)
        dx = targets[k].x - brickpos[i].x;
      else if (brickpos[i].x + bricks[i].size - 1 < targets[k].x)
        dx = targets[k].x - brickpos[i].x - bricks[i].size + 1;
      else
        dx = 0;
      dy=abs(brickpos[i].y - targets[k].y);
      d = dx * dx + dy * dy;
      if (d < target_dist[k])
        target_dist[k] = d;
    }

  alltargets=1;
  for (k=0;k<target_count;k++) {
    if (target_dist[k] == 0) fit += 1.0;
    else {
      alltargets=0;
      target_dist[k] = sqrt(target_dist[k]);
      fit +=exp((double) (-target_dist[k]) / 100.0 ) ;
    }
  }

  if ((alltargets) && (load_count>0)) {
    forcecount++;
    for (k=0;k<load_count;k++) {
      i=getgrid(loads[k].x,loads[k].y);
#ifdef CHECKS
      if ((!(ISSOLID(i))) || (i >= bodycount))
        abortonerror(342,"fitness: wrong attach %d at (%d,%d) bc=%d",
                     i,loads[k].x,loads[k].y,bodycount);
#endif
      forcelist[forcecount-1].target=i;
      forcelist[forcecount-1].x.x = ((REAL) loads[k].x) + .05;
      forcelist[forcecount-1].x.y = ((REAL) loads[k].y + 0.5)* HEIGHT ;
      forcelist[forcecount-1].f = mul2d(loads[k].load,grav2d);
    }
    netvalue=flow(bodies,bodycount,joints,jointcount,
                  forcelist,forcecount);
    if (netvalue > 1.0) netvalue=1.0;
    fit += netvalue + avgload/100;
  }

  /* fit = fit / ((REAL) (1 + target_count)); */
  fit = fit / (double)(1 + target_count + load_count);
  /*
  if (fit == 0.0) {
    printf("ZERO xFITNESS!\n");
    abortonerror(238,"zero fitness");
  }
  */

  return fit;
}
