/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#include "headers.h"


void unstabi_up(body *b) {
  int j; body *son; joint *J;
  if (b==NULL) return;

#ifdef PRINTS
  printf("unstabi_up %d\n",b->n);
#endif
  b->rooted = FALSE;
  b->equilibrium = FALSE;
  for (j=0;j<b->jointcount;j++) {
    son = b->joins[j];
#ifdef PRINTS
    printf("(son %d)",son==NULL ? -1:son->n);
#endif
    J = b->joints[j];
    if (son != NULL &&
	son->equilibrium == TRUE &&
	fabs(J->torques / J->maxtorque) < 0.999
	) unstabi_up(son);
#ifdef PRINTS
    else printf("no pass\n");
#endif
  }
  b->equilibrium = TRUE;
}


int find_falling(int init) {
  int i,j;
  int prev_mode = simulation_mode;
  int ret=0;

  simulation_mode = MULTITEST_MODE;
  life_count=0;

  for (i=0;i<obj_count;i++) {
    if (objects[i].kind == BRICK_KIND) {
      if (init)
	still_alive[i]=1;
      if (still_alive[i])
	life_count++;
    }
    else
      still_alive[i]=0;
  }

  if ((life_count <= 0) || object_kind_count[GROUND_KIND] <= 0) {
      /* trivial problem, nothing's falling 'cause there's notin' there */
    ret = 0;
    goto end;
  }
  do {
    static int count=0;
#ifdef PRINTS
    char fname[30];
#endif
    count++;

#ifdef PRINTS
    printf("LOOP ITER %d\n",count);
#endif

    if (make_object_model_conditional())  {
#ifdef PRINTS
      fprintf(stderr,"makemodel failed!\n");
#endif
#if defined(DEBUG) || defined(PRINTS)
      load_objects_conditional();
      mat_bricks("tmp.m");
#endif
      ret = -1;
      goto end;
    }
    if ((j=tooheavy_search())) {
#ifdef PRINTS
      fprintf(stderr,"tooheavy_search failed!\n");
#endif

#if defined(DEBUG) || defined(PRINTS)
      sprintf(fname,"fail%d.m",count);
      mat_bricks(fname);
#endif
      brickcount=0;
      for (i=0;i<obj_count;i++) {
	if (still_alive[i]) {
	  if (!bodies[brickcount].equilibrium) {
	    still_alive[i]=0;
	    life_count --;
	  }
	  brickcount++;
	}
      }
    }
  } while (j && (life_count > 0));

  /* yipee! we've found a stable substructure */
  ret=0;
 end:
  simulation_mode=prev_mode;
  return ret;
}

/******************/
/* multi version  */
/******************/


int tooheavy_search() {
  /* init joints and bodies */
  int i;
  int ret;
  REAL netvalue;
  problem_type *p;

  for (i=0; i<bodycount ;i++)
    bodies [i].equilibrium = 1;
  for (i=0;i<jointcount ;i++) {
    joints  [i].torques = 0.0;
    joints  [i].forces.x = joints  [i].forces.y = 0.0;
  }
  addforcecount	 =MAXADDFORCECOUNT;
  ret = 0;
  for (i=0;i<bodycount ;i++) {
    forcelist[i].target = i;
    forcelist[i].x = bodies [i].x;
    forcelist[i].f = mul2d(bodies [i].mass,grav2d);
  }
  forcecount=bodycount;
  { /* add loads */
    int i,k;
    for (k=0;k<obj_count;k++) {
      if (objects[k].kind == LOAD_KIND) {
	i=getgrid(objects[k].x,objects[k].y);
	if (i>=0) {
	  forcelist[forcecount-1].target=i;
	  forcelist[forcecount-1].x.x =
	    ((REAL) objects[k].x) + .05;
	  forcelist[forcecount-1].x.y =
	    ((REAL) objects[k].y + 0.5)* HEIGHT ;
	  forcelist[forcecount-1].f =
	    mul2d(objects[k].size / WEIGHT,grav2d);
	  forcecount++;
	}
      }
    }
  }

#ifdef DEBUG
  save_file_after_parse=1;
#endif
  netvalue = multi(bodies,bodycount,joints,jointcount,
		   forcelist,forcecount, &p);

  if (netvalue == 1.0)
    ret =0; /* structure stands. no more calculations needed */
  else {
    /* it doesn't stand, search for who's falling! */
    netvalue = itermultitest (bodies , bodycount ,
		   joints , jointcount ,
		   forcelist , forcecount , &p);

    ret = 0;
    for (i=0;i<forcecount;i++) {
      if ( p->flow[p->n[0]+((1+p->n[0]) * i)] < 0.98) {
	ret = 1;
	break;
      }
    }
  }
  if (ret) {
    /* unparse to search for falling bricks */
    unparse_3(p,bodies,bodycount,
	      joints,jointcount,
	      forcelist,forcecount);

    find_unstables();

#ifdef PRINTS
    printf("Too heavy.\n");
#endif
  }
  return ret;
}

void find_unstables() {
  /* for multiflow only */
  int i;
  int arcno;
  int target;
  problem_type *p = the_problem;

  for (i=0;i<bodycount;i++) {
    bodies[i].rooted=TRUE;
    bodies[i].equilibrium=TRUE;
    /* equilibrium is only being used as a marker here,
       to prevent recursive calls */
  }
  for (i=0;i<forcecount;i++) {
    arcno = p->n[0]; /* last arc = arc that connects
			ground to sink */
    if (p->flow[arcno+i*(1+p->n[0])] < 0.99) {
      /* we've found an unstable brick */
      /* propagate unstability */
      target = forcelist[i].target; /* this is the brick where the force was applied */
      unstabi_up(&bodies[target]);
    }
  }
  for (i=0;i<bodycount;i++)
    bodies[i].equilibrium = bodies[i].rooted;
}


int convert_problem() {
  /* we wish to adapt a list of objects and convert it into a
     stable structure to seed the population. Stable, i.e., it's not
     falling according to the fast simulation. So what we do, we use
     itermultitest to find the best flow possible, then use the flow
     directions to determine paternity relationships. Perhaps even
     joint's weights. The we prune the tree until the thing is
     stable */

  int i,j;
  int prev_mode = simulation_mode;
  int ret=0;

  simulation_mode = MULTITEST_MODE;
  fprintf(stderr,"Enter convert_problem\n");

  /* all are alive so far */
  for (i=0;i<obj_count;i++) {
    if (objects[i].kind == BRICK_KIND) {
      still_alive[i]=1;
      life_count++;
    }
    else
      still_alive[i]=0;
  }

  if ((object_kind_count[GROUND_KIND]) <= 0) {
    abortonerror(621,"No grounds!");
    /* can't evolve without grounds */
    /* ret=-1; goto end; */
  }

  if (life_count <= 0) {
    /* trivial problem, nothing's falling 'cause there's notin' there */
    ret = 0;
    goto end;
  }

  if (make_object_model_conditional())	{
#ifdef PRINTS
    fprintf(stderr,"makemodel failed!\n");
#endif
#if defined(DEBUG) || defined(PRINTS)
    load_objects_conditional();
    mat_bricks("tmp.m");
#endif
    abortonerror(622,"momc failed@convert");
    /*	    ret = -1;  goto end; */
  }


  for (i=0; i<bodycount ;i++)
    bodies [i].equilibrium = 1;
  for (i=0;i<jointcount ;i++) {
    joints  [i].torques = 0.0;
    joints  [i].forces.x = joints  [i].forces.y = 0.0;
  }
  addforcecount = MAXADDFORCECOUNT;
  ret = 0;
  for (i=0;i<bodycount ;i++) {
    forcelist[i].target = i;
    forcelist[i].x = bodies [i].x;
    forcelist[i].f = mul2d(bodies [i].mass,grav2d);
  }

  {

    double netvalue;
    problem_type *p;
    int arcno,j,inverse_arc;
    int head,tail;
    double flow,invflow,one;
    int *arc_seen;

#ifdef DEBUG
    save_file_after_parse=1;
#endif
    netvalue = itermultitest (bodies , bodycount ,
		   joints , jointcount ,
		   forcelist , bodycount , &p);

    fprintf(stderr,"itermultitest returned %f\n",netvalue);
    /* we ignore the netvalue returned,
       for there's nothing special about the last node */

    /* for each arc: compare flow with reverse flow
       and decide wich way the flow flows...
       perhaps smarter would be to add torques not flows
    */
    ARRALLOC(arc_seen,p->n[0]);
    for (arcno=0;arcno< (p->n[0]);arcno++) arc_seen[arcno]=0;

    for (arcno=0;arcno < (p->n[0]) /*no of arcs */; arcno++) {
      if (arc_seen[arcno]) continue;
      head = p->oa[arcno] - 1;
      tail = p->da[arcno] - 1;
      fprintf(stderr,"arc %d %d->%d bodycount %d\n",arcno,head,tail,bodycount);
      if ((p->da[arcno]  <= bodycount)
	  && (p->oa[arcno] <= bodycount) ) {
	/* this arc is a brick-brick joint */
	inverse_arc = -1;
	for (j=arcno+1;j < (p->n[0]);j++) {
	  if (p->oa[arcno] == p->da[j]
	      && p->da[arcno] == p->oa[j]) {
	    fprintf(stderr,"found inverse of arc %d = %d\n",arcno,j);
	    inverse_arc = j;
	    break;
	  }
	}
	if (inverse_arc >=0) arc_seen[inverse_arc] = 1;

	flow=invflow=0.0;
	/* add total */
	for (i=0;i<(p->k[0]);i++) /* for each commodity */ {
	  one = p->flow[1+arcno + (p->n[0]+1)*(i)];
	  flow += one;
	  /* fprintf(stderr,"flow of %d through %d = %f\n",i,arcno,one);  */
	  if (inverse_arc >=0) {
	    one = p->flow[1+inverse_arc + (p->n[0]+1)*(i)];
	    invflow += one;
	    /* fprintf(stderr,"flow of %d through %d = %f\n",i,inverse_arc,one); */
	  }
	}
	if (0.0 != invflow && 0.0 != flow) {
	  fprintf(stderr,"warning: double flow %d -> %d (%f,%f)\n",
		  head,tail,flow,invflow);
	}
	if (flow >= invflow) i=arcno;
	else {
	  i=inverse_arc;
	  j=tail; tail=head; head=j;
	}
	/* now we know that weight flows FROM head to tail,
	   so tail is head's parent */
	if (parent[head] != GROUND) {
	  fprintf(stderr,"Head %d already has parent %d not %d!!\n",
		  head,parent[head],tail);
	}
	/* fprintf(stderr,"%d adopts %d \n",tail,head); */
	adopt_brick(head,parent[head],tail,0);
      }
    }
    free(arc_seen);
  }

  {
    /* now we've built a tree with a minimum number of roots
       let's remove random leaves, as necessary, until we
       get it to be stable  (I know, we could be smarter about
       this... */

    int heavy;
    model_type model;

    simulation_mode=EMBEDDED_MODE;
    ARRALLOC(model.bricks,MAXBRICKS);
    ARRALLOC(model.roots,MAXROOTS);

  copy_to_my_model:

    for (i=0;i<brickcount;i++) {
      model.bricks[i] = bricks[i];
    }
    for (i=0;i<rootcount;i++) {
      model.roots[i]=roots[i];
    }
    model.rootcount=rootcount;
    model.brickcount=brickcount;

  load_my_model:
    loadmodel(model.roots,model.rootcount,model.bricks);
    fprintf(stderr,"model roots %d %d brickcount %d %d\n",
	    model.rootcount,rootcount,model.brickcount,brickcount);
    makemodel();
    /* delete all floating roots */
    for (i=0;i<brickcount;i++) {
      if (bodies[i].jointcount == 0) {
	fprintf(stderr,"body %d has no joints\n",i);
	for (j=0;j<rootcount;j++) {
	  if (roots[j].no == i) break;
	}
	if (j == rootcount)
	  abortonerror(424,"floater %d is nonroot",i);
	for (i=j+1;i<rootcount;i++) {
	  roots[i-1]=roots[i];
	}
	rootcount--;
	goto copy_to_my_model;
      }
    }

    heavy=tooheavy();
    if (heavy && brickcount > 0) {
      /* delete a random leaf */
      fprintf(stderr,"Too heavy, killing random leaf\n");
      for (i=0;i<brickcount;i++) model.bricks[i] = bricks[i];
      for (i=0;i<rootcount;i++) model.roots[i] = roots[i];
      model.rootcount=rootcount;
      model.brickcount=brickcount;
      if (kill_brick(model.roots,&model.rootcount,
		     model.bricks,model.brickcount,
		     (RANDINT() % brickcount)))
	abortonerror(532,"kill_brick misteriously failed");
      goto load_my_model;
    }
    free (model.bricks);
    free (model.roots);
  }
  /* yipee! we've found a stable thing */
  ret=0;

 end:
  fprintf(stderr,"convert_model exits (%d)\n",ret);
  simulation_mode=prev_mode;
  return ret;
}
