/*---------------------------------------------------------------
This file is part of the Buildable Evolution Software Project.

http://www.demo.cs.brandeis.edu/pr/buildable
(C) Pablo Funes and Jordan B. Pollack 1997-2008
pablo@cs.brandeis.edu

The methods and configurations incorporated in the Builable Evolution
Software Project are covered in part by United States patent
6,434,492 "Computer apparatus and method for analyzing structural
stability" of August 13, 2002.

The Buildable Evolution Software Project is free software: you
can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option)
any later version.

The Buildable Evolution Software Project is distributed in the
hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License
along with the software. If not, see <http://www.gnu.org/licenses/>.
---------------------------------------------------------------*/
#define SIMUL_C

#include "headers.h"


/* addforce abortion mechanism */
int addforcecount OFE;

/*#define WEAKNESS 0.8 */

#ifdef USEALARM
#include <signal.h>
#include <unistd.h>
#endif

brick bricks OFE [MAXBRICKS];
xybrickptr roots [MAXROOTS];
int rootcount;

joint joints OFE [MAXJOINTS];
body bodies OFE [MAXBODIES];

REAL totalmass OFE;
int brickcount OFE;
int jointcount OFE;
int bodycount OFE;
REAL fitnessvalue OFE;

#define MAXFORCES (MAXBODIES+1)
struct appliedforce forcelist OFE [MAXFORCES];


short parent OFE [MAXBRICKS];

int adoptcount OFE;

brick newbrick(int size) {
  int j;
  brick b;
  b.size = size;
  for (j=0;j<BRICKJOINTS;j++) {
    b.joints[j] = EMPTY;
    b.jointpos[j].x =
      b.jointpos[j].y = 0;
    OBJCPY(b.flow,evenflow);
  }
  return b;
}


int free_joint(brick *b) {
  int j,ret;
  ret=-1;
  for (j=0;j<BRICKJOINTS;j++)
    if (b->joints[j] == EMPTY) {
      ret=j;
      break;
    }
  return ret;
}

void makejoint(int body1, int body2, int sign, int jointsize,
	       REAL x, REAL y,REAL weight) {
  USEMINE
  joint *j;
#ifdef PRINTS
  printf("makejoint %d %d %d %d %f %f\n",body1,body2,sign,jointsize,x,y);
#endif
  /* adopt if needed */
  if (parent[body1] != body2
      && adoptcount < MAXBRICKS
      && (body2 >= 0 || body2 == GROUND)
      && body1 >= 0) {
    adoptives[adoptcount].son = body1;
    adoptives[adoptcount].father = body2;
    adoptcount++;
  }
  j=&joints MINE [jointcount MINE ++];
  if (jointcount MINE == MAXJOINTS) {
    fprintf( stderr,"Error: Too many joints!\n");
    abortonerror(132,"too many joints");
  }
  j->x.x = x;
  j->x.y = y;
  j->weight = weight;
  if (jointsize < 0) {
   abortonerror(343,"negative jointsize");
  }
  switch(jointsize) {

  case 0: j->maxtorque = 0.0; break;
  case 1: j->maxtorque = 405.0; break;
  case 2: j->maxtorque = 1960.0; break;
  case 3: j->maxtorque = 3500.0; break;
  case 4: j->maxtorque = 6144.0 /*9600.0*/; break;
  case 5: j->maxtorque = 11000.0 /* 12672.0 */; break;
  case 6: j->maxtorque = 13520.0; break;
  case 7: j->maxtorque = 14238.0; break;
  default:j->maxtorque = jointsize * 14238.0 / 7.0;
  }
  /* j->maxtorque = jointsize * TORQ_UNIT */;
  j->n = jointcount MINE;
  addjoint(&bodies MINE [body1], sign,
	   ((body2==GROUND) ? NULL : &bodies MINE[body2]),j);

  if (simulation_mode != EMBEDDED_MODE)
    if (body2 != GROUND)
      addjoint(&bodies MINE [body2], -sign, &bodies MINE [body1],j);
}

int newjoints(int gx, int gy, int bodyno, int size, int sign) {
  int k,found,x,at;
  body *body;
  body = &bodies MINE [bodyno];
  found=EMPTY;at=-1000;
  for (k=0;k<size;k++,gx++) {
    x=getgrid(gx,gy+sign);
    if (found != EMPTY && x != found && ISSOLID(found)) {
      makejoint(bodyno,
		found,sign, gx - at,
		((REAL) at + gx) / 2.0,
		(gy + 0.5 + 0.5 * sign) * HEIGHT,
		bricks[body->n].flow[body->jointcount]);
       at = gx;
    }
    if (found == EMPTY) at = gx;
    found=x;
  }
  if (ISSOLID(found))
    makejoint(bodyno,found,sign,gx-at,
	      ((REAL) at+ gx) / 2.0,
	      (gy + 0.5 + 0.5 * sign) * HEIGHT,
	      bricks[body->n].flow[body->jointcount]);
  return 0;      /* never fails */
}


int putbrickat(int gx, int gy, int size,int brickno) {
  USEMINE
  int j;

  /* or out of bounds in general */
  if (gy >= maxy || gy < miny || gx < minx || gx+size >= maxx)
    return 1;

  /* see if this brick fits */
  for (j=0;j<size;j++)
    if (EMPTY != getgrid(gx+j,gy))
      return 1;

  /* does fit, add it to the grid */
   for (j=0;j<size;j++)
     setgrid(gx+j,gy,brickno/* bodycount MINE was an error? */);

  return 0;
}

#define putnewbrickat(x,y,s) putbrickat(x,y,s,brickcount)

int addbrick(short gx, short gy, short brickno, short father) {
  USEMINE
  int j,k;

  #ifdef PRINTS
  printf("AB %d %d %d(%d) ",gx,gy,brickno,bricks MINE [brickno].size);
#endif

 if (bodycount MINE == MAXBODIES) return 1;    /* failed: too many bodies */

 /* see if this brick fits and add it to the grid */
 if (putbrickat(gx,gy,bricks MINE [brickno].size,brickno))
   return 1; /* if it don't fit, you must acquit */

#ifdef CHECKS
 if (brickno != bodycount) {
   abortonerror(875,"brickno %d != bodycount %d@ addbrick",brickno,bodycount);
 }
#endif

 brickpos[brickno].x = gx;
 brickpos[brickno].y = gy;
 parent[brickno] = father;

 /* create body */
 bodies MINE [bodycount MINE ].x.x = 0.5 * bricks MINE[brickno].size + gx;
 bodies MINE [bodycount MINE ].x.y = HEIGHT * (gy + 0.5);
 bodies MINE [bodycount MINE ].mass = bricks MINE[brickno].size;
 bodies MINE [bodycount MINE ].jointcount = 0;
 bodies MINE [bodycount MINE] .n = brickno;
 totalmass MINE += bodies MINE [bodycount MINE].mass;


 /* create joints */
 if (newjoints(gx,gy,bodycount MINE,bricks MINE[brickno].size,-1) ||
     newjoints(gx,gy,bodycount MINE,bricks MINE[brickno].size,1))
   return 1;                   /* failed */

  bodycount MINE++;

  /* create linked bricks */
  for (j=0;j<BRICKJOINTS;j++) {
    if ((k=bricks MINE[brickno].joints[j]), ISBRICK(k)) {
#ifdef PRINTS
      printf("J %d(%d %d) k=%d(%d) ",j,bricks MINE[brickno].jointpos[j].x,
	     bricks MINE[brickno].jointpos[j].y,k, bricks MINE[k].size);
#endif
      if (addbrick (
		    gx + bricks MINE [brickno].jointpos[j].x,
		    gy + bricks MINE [brickno].jointpos[j].y,
		    k,brickno))
	return 1;    /* failed */
    }
  }
  return 0;
}

int makemodel() {
  USEMINE
  int ret;
  int i;
  int c;
#ifdef PRINTS
  printf("MM ");
#endif
  initgrid();
  ret=0;
  jointcount MINE = 0;
  bodycount MINE = 0;
  totalmass MINE= 0.0;
  adoptcount = 0;
  for (i=0;i<rootcount;i++)
    if ((ret=addbrick(roots[i].x,roots[i].y,roots[i].no,-1)) != 0 )
      break;
#ifdef PRINTS
  printf("ab000=%d ",ret);
  #endif
  if (ret==0) {
    /* add predefined bricks */
    c=brickcount MINE;
    reserve_spots(EMPTY); /* de-regulate previously reserved spots */
    if (add_predefs(c)) ret=1;
  }
#ifdef PRINTS
  if (ret) printf ("MM Failed\n");
  else printf("\n");
#endif
  return ret;
}


void mtree(FILE *fp,int gx, int gy, int brickno, brick bricks[]) {
  brick *b;
  int j,k;

  b=&bricks[brickno];
  fprintf(fp,"brick2d([%g %g],[%g %g],%d,%d);\n",
          /*x,y */ (REAL) gx, (REAL) gy*HEIGHT,
          /* dx, dy */ (REAL) bricks[brickno].size, HEIGHT,
          /*brickno, bricksize */ brickno,bricks[brickno].size);

  for (j=0;j<4;j++) {
    k=b->joints[j];
    if (k != EMPTY && k != GROUND) {
      mtree(fp,
            gx + bricks[brickno].jointpos[j].x,
            gy + bricks[brickno].jointpos[j].y,
            k,bricks);
      /*
      fprintf(fp,"text(%f,%f,'%d,%d');\n",
              (REAL) gx + ((j==0 || j==3) ? 0 : bricks[brickno].size - 2),
              (REAL) HEIGHT * (gy + ((j==0 || j==1) ? 0 : 1)),
              bricks[brickno].jointpos[j].x,
              bricks[brickno].jointpos[j].y); */
    }

  }
}

void mmbricks(void) {
  mat_bricks("tmp.m");
}

void mbricks(char *name,
             xybrickptr roots[], int rootcount,
             brick bricks[],int brickno,
             double fit,
             body bodies[], int bodycount,
             joint joints[], int jointcount) {
  FILE *fp;
  body *b;
  int i,j;

  brickno++;

  fp=fopen(name,"w");
  fprintf(fp,"figure; hold on\n");
  for (i=0;i<rootcount;i++)
    mtree(fp,roots[i].x,roots[i].y,roots[i].no,bricks);
  for (i=0;i<bodycount;i++) {
    b = &bodies [i];
    /*    fprintf(fp,"plot(%f,%f,'co');\n",b->x.x,b->x.y); */
    for (j=0;j < b->jointcount;j++) {
      fprintf(fp,"joint2d([%g %g],[%g %g],",
              /*x,y*/ b->joints[j]->x.x, b->joints[j]->x.y,
              /* from */ b->x.x, b->x.y);
      if (b->joins[j] == NULL) {
        /* ground joint */ fprintf(fp,"[],");
      }
      else {
        fprintf(fp,"[%g %g],",
                /* to (x,y) */ b->joins[j]->x.x, b->joins[j]->x.y);
      }
      fprintf(fp,"%g,%g);\n",
              b->joints[j]->torques,
              b->joints[j]->maxtorque);
    }
  }
  for (i=0;i<obj_count;i++) {
    REAL x,y,xo,yo,size;
    x=(REAL) objects[i].x;
    y=(REAL)objects[i].y * HEIGHT;
    xo=x+0.5;
    yo=y+HEIGHT/2;
    size=(REAL) objects[i].size;
    switch (objects[i].kind) {
    case GROUND_KIND:
      fprintf(fp,"ground2d([%g %g],[%g %g],%d,%d);\n",
	      x,y,size,HEIGHT,i,objects[i].size);

      break;
    case TARGET_KIND:
      fprintf(fp,"target2d([%g %g]);\n",xo,yo);
      break;
    case LOAD_KIND:
      fprintf(fp,"load2d([%g %g],%g);\n",xo,yo,size);
      break;
    case RESTRICTION_KIND:
      fprintf(fp,"restr2d([%g %g]);\n",xo,yo);
    case FIXED_KIND:
      fprintf(fp,"brick2d([%g %g],[%g %g],%d,%d);\n",
	      x,y,size,HEIGHT,i,objects[i].size);
      break;
    }
  }
  fprintf(fp,"title('%f')\n",fit);
  fclose(fp);
}

void mat_bricks(char *name) {
  mbricks(name,
          roots,rootcount,
          bricks,brickcount,
          fitnessvalue,
          bodies,bodycount,
          joints,jointcount);
}



#ifdef VBLOCKS
void vb_tree(FILE *fp,int gx, int gy, int brickno) {
  brick *b;
  int j,k;

  b=&bricks[brickno];
  fprintf(fp,"%d     %d %d    %d %d %d       %d %d %d\n",
	  1,
	  1,b->size,
	  0,(gy-4)*3,-(gx+b->size),
	  0,0,0);
  for (j=0;j<4;j++) {
    k=b->joints[j];
    if (k != EMPTY && k != GROUND) {
      vb_tree(fp,
	    gx + bricks[brickno].jointpos[j].x,
	      gy + bricks[brickno].jointpos[j].y,k);

    }
  }
}

vb_bricks(char *name,brick bricks[],int brickno,double fit) {
  FILE *fp;
  body *b;
  int i,j;
  int gx,gy;

  brickno++;

  fp=fopen(name,"w");
  fprintf(fp,"2 	10 40 	-5 -13 -40	 0 	7 	0\n");
  vb_tree(fp,brick0x,brick0y,0);
  fclose(fp);
}


#endif

void dumpfile(char *name,
	      xybrickptr roots[],int rootcount,
	      brick bricks[], int brickcount,
	      body bodies[], int bodycount,
	      joint joints[], int jointcount) {
  FILE *fp;
  int i,j;

  fp=fopen(name,"w");
  fprintf(fp,"-1\n");
  fprintf(fp,"\n%d\n",rootcount);
  for (i=0;i<rootcount;i++)
    fprintf(fp,"%d %d %d\n",roots[i].x,roots[i].y,roots[i].no);
  fprintf(fp,"\n%d\n",brickcount);
  for (i=0;i<brickcount;i++) {
    fprintf(fp,"%5d %d ",i,bricks[i].size);
    for (j=0;j<4;j++) {
      fprintf(fp,"%d %d,%d   ",bricks[i].joints[j],bricks[i].jointpos[j].x,
	      bricks[i].jointpos[j].y);
    }
    fprintf(fp,"\n\n");
  }

  for (i=0;i<brickcount;i++) {
    for (j=0;j<BODYJOINTS;j++) {
      fprintf(fp,"%g ",bricks[i].flow[j]);
    }
    fprintf(fp,"\n");
  }

  fprintf(fp,"\n\n\n---------bodies(%d)-------\n\n",bodycount);
  for (i=0;i<bodycount ;i++) {
    fprintf(fp,"%5d=%8p (%f,%f) eq %d mass %f ",i,&bodies[i],
	   bodies[i].x.x,
           bodies[i].x.y,
	    bodies[i].equilibrium,
	   bodies[i].mass);
    for (j=0;j<bodies[i].jointcount;j++)
      fprintf(fp,"%p(%p %d) ",bodies[i].joins[j],
	     bodies[i].joints[j],bodies[i].jointsign[j]);
    fprintf(fp,"\n");
  }

  fprintf(fp,"\n\n-----------joints(%d)----------\n\n",jointcount);
  for (i=0;i<jointcount;i++) {
    fprintf(fp,"%5d=%8p (%f,%f) t=%f f=%f,%f, mt=%f\n",
	   i,
	   &joints[i],
	   joints[i].x.x,joints[i].x.y,
	   joints[i].torques,joints[i].forces.x,joints[i].forces.y,
	   joints[i].maxtorque);
  }

  fclose(fp);
}

void dump_bricks(char *name) {
  dumpfile(name,roots,rootcount,bricks,brickcount,bodies,bodycount,joints,jointcount);
}
