/* reals and booleans */

#define REAL float

#define BOOL int
#define TRUE 1
#define FALSE 0

#define NOTHING 0.000001
#define MIN(a,b) ((a) > (b)) ? (b) : (a)
#define MAX(a,b) ((a) < (b)) ? (b) : (a)
#define SIGN(a) (((a) == 0) ? 0 : (((a) > 0) ? 1 : -1))

/* 2d vectors and their operations */

typedef struct {
  REAL x,y;
} vector2d;


vector2d v2d(REAL x, REAL y);

REAL reallyzero(REAL x) ;

REAL norm2d(vector2d v) ;

REAL dot2d(vector2d v, vector2d w) ; 

vector2d sum2d(vector2d v, vector2d w);


vector2d mul2d(REAL a, vector2d v); 

vector2d sub2d(vector2d v, vector2d w);

REAL dist2d(vector2d v, vector2d w);

REAL det2d(vector2d v, vector2d w) ;

BOOL zero2d(vector2d v) ;
BOOL reallyzero2d(vector2d v) ;

BOOL parallel2d(vector2d v, vector2d w) ;

/* forces and torques */
typedef vector2d force2d;

typedef struct {
  force2d f;
  vector2d x;
} relforce;

typedef REAL torque2d;

torque2d torqueat(vector2d mass, vector2d x, force2d f) ;

/* bodies and joints */

#define BODYJOINTS 10

typedef struct {
  vector2d x;
  torque2d torques;
  force2d forces;
  torque2d maxtorque;
  float weight;
  int n;
} joint;

typedef struct body_struct{
  vector2d x;
  BOOL equilibrium;
  BOOL rooted;
  joint *joints[BODYJOINTS];
  struct body_struct  *joins[BODYJOINTS];
  int jointflow[BODYJOINTS];
  int jointsign[BODYJOINTS];
  int jointcount;
  REAL mass;
  int n;
} body ;


REAL addforce(body *b, vector2d x, force2d f) ;


void addjoint(body *b1, int sign, body *b2, joint *j) ;














